// Copyright (c) 1995 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif

#include "config.h"
#include "Boolean.H"
#include "EasyParser.H"
#include "Event.H"
#include "macros.H"
#include "SGMLApplication.H"
#include "ParserEventGenerator.H"

struct ParserEventGeneratorImpl {
  EasyParser easy;
  unsigned nErrors;
  bool messagesInhibited;
};

class SpOpenEntity : public SGMLApplication::OpenEntity {
public:
  SpOpenEntity(const ConstResourcePointer<Origin> &origin);
  SGMLApplication::Location location(SGMLApplication::Position) const;
private:
  ConstResourcePointer<Origin> origin_;
};

class AppEventHandler : public EventHandler {
public:
  AppEventHandler(SGMLApplication *, MessageReporter *reporter,
		  const bool *messagesInhibitedPtr);
  ~AppEventHandler();
  unsigned nErrors() const;
  void appinfo(AppinfoEvent *);
  void startDtd(StartDtdEvent *);
  void endDtd(EndDtdEvent *);
  void endProlog(EndPrologEvent *);
  void startElement(StartElementEvent *);
  void endElement(EndElementEvent *);
  void data(DataEvent *);
  void pi(PiEvent *);
  void sdataEntity(SdataEntityEvent *);
  void externalDataEntity(ExternalDataEntityEvent *);
  void subdocEntity(SubdocEntityEvent *);
  void commentDecl(CommentDeclEvent *);
  void ignoredChars(IgnoredCharsEvent *);
  void markedSectionStart(MarkedSectionStartEvent *);
  void markedSectionEnd(MarkedSectionEndEvent *);
  void message(MessageEvent *);
  void *allocate(size_t);
  void freeAll();
  void freeAll1();

  struct Block {
    Block *next;
    char *mem;
    size_t size;
  };
    
  static void setString(SGMLApplication::CharString &, const CString &);
  static void clearString(SGMLApplication::CharString &);
private:
  AppEventHandler(const AppEventHandler &); // undefined
  void operator=(const AppEventHandler &);  // undefined
  void setLocation(SGMLApplication::Position &, const Location &);
  void setLocation1(SGMLApplication::Position &, const Location &);
  void setAttributes(const SGMLApplication::Attribute *&attributes,
		     const AttributeList &attributeList);
  void setExternalId(SGMLApplication::ExternalId &to,
		     const ExternalId &from);
  void setEntity(SGMLApplication::Entity &to, const Entity &from);
  void setNotation(SGMLApplication::Notation &, const Notation &notation);
  static void clearNotation(SGMLApplication::Notation &);
  static void clearExternalId(SGMLApplication::ExternalId &);
  SGMLApplication *app_;
  MessageReporter *reporter_;
  unsigned nErrors_;
  ConstResourcePointer<Origin> lastOrigin_;
  SGMLApplication::OpenEntityPtr openEntityPtr_;
  size_t firstBlockUsed_;
  size_t firstBlockSpare_;
  Block *freeBlocks_;
  Block *allocBlocks_;
  const bool *messagesInhibitedPtr_;
};

inline
void *operator new(size_t n, AppEventHandler *handler)
{
  return handler->allocate(n);
}

inline
unsigned AppEventHandler::nErrors() const
{
  return nErrors_;
}

inline
void AppEventHandler::freeAll()
{
  if (allocBlocks_)
    freeAll1();
}

inline void
AppEventHandler::setString(SGMLApplication::CharString &to,
			   const CString &from)
{
  to.ptr = from.pointer();
  to.len = from.length();
}

inline
void AppEventHandler::clearString(SGMLApplication::CharString &to)
{
  to.len = 0;
}

inline
void AppEventHandler::clearNotation(SGMLApplication::Notation &to)
{
  clearString(to.name);
}

inline
void AppEventHandler::setLocation(SGMLApplication::Position &pos,
				  const Location &loc)
{
  if (lastOrigin_ != loc.origin())
    setLocation1(pos, loc);
  else
    pos = loc.index();
}

ParserEventGenerator::ParserEventGenerator()
{
  impl_ = new ParserEventGeneratorImpl;
  impl_->nErrors = 0;
  impl_->messagesInhibited = 0;
}

void ParserEventGenerator::inhibitMessages(bool b)
{
  impl_->messagesInhibited = b;
}

ParserEventGenerator::~ParserEventGenerator()
{
  delete impl_;
}

void ParserEventGenerator::init(int nFiles, char *const *files)
{
  impl_->easy.init(nFiles, files);
}

void ParserEventGenerator::run(SGMLApplication &app)
{
  AppEventHandler handler(&app, &impl_->easy, &impl_->messagesInhibited);
  impl_->easy.parser().parseAll(handler);
  impl_->nErrors = handler.nErrors();
}

unsigned ParserEventGenerator::nErrors() const
{
  return impl_->nErrors;
}

void ParserEventGenerator::setOption(Option opt)
{
  switch (opt) {
  case warnDuplicateEntity:
    impl_->easy.options.warnDuplicateEntity = 1;
    break;
  case warnDefaultEntity:
    impl_->easy.options.warnDefaultEntityReference = 1;
    break;
  case warnUndefinedElement:
    impl_->easy.options.warnUndefinedElement = 1;
    break;
  case showOpenEntities:
    impl_->easy.addOption(MessageReporter::openEntities);
    break;
  case showOpenElements:
    impl_->easy.addOption(MessageReporter::openElements);
    break;
  case outputCommentDecls:
    impl_->easy.options.eventsWanted.addCommentDecls();
    break;
  case outputMarkedSections:
    impl_->easy.options.eventsWanted.addMarkedSections();
    break;
  }
}

void ParserEventGenerator::setOption(OptionWithArg opt, const char *arg)
{
  switch (opt) {
  case addCatalog:
    impl_->easy.addCatalogSysid(arg);
    break;
  case includeParam:
    impl_->easy.addInclude(arg);
    break;
  }
}

AppEventHandler::AppEventHandler(SGMLApplication *app,
				 MessageReporter *reporter,
				 const bool *messagesInhibitedPtr)
: app_(app), nErrors_(0), reporter_(reporter),
  messagesInhibitedPtr_(messagesInhibitedPtr),
  freeBlocks_(0), allocBlocks_(0), firstBlockSpare_(0), firstBlockUsed_(0)
{
}

AppEventHandler::~AppEventHandler()
{
  freeAll();
  while (freeBlocks_) {
    Block *tem = freeBlocks_;
    freeBlocks_ = freeBlocks_->next;
    delete [] tem->mem;
    delete tem;
  }
}

void AppEventHandler::freeAll1()
{
  Block **p;
  for (p = &allocBlocks_; *p; p = &(*p)->next)
    ;
  *p = freeBlocks_;
  freeBlocks_ = allocBlocks_;
  allocBlocks_ = 0;
}

void *AppEventHandler::allocate(size_t n)
{
  if (n == 0)
    return 0;
  // round up to avoid alignment problems
  n = (n + sizeof(char *) - 1) & ~(sizeof(char *) - 1);
  enum { BIG = 1024 };
  if (n > firstBlockSpare_) {
    if (freeBlocks_ && firstBlockSpare_ < BIG) {
      Block *tem = freeBlocks_;
      freeBlocks_ = freeBlocks_->next;
      tem->next = allocBlocks_;
      allocBlocks_ = tem;
    }
    if (!freeBlocks_ || freeBlocks_->size < n) {
      Block *tem = new Block;
      tem->size = n < BIG ? int(BIG) : n;
      tem->mem = new char[tem->size];
      tem->next = freeBlocks_;
      freeBlocks_ = tem;
    }
    firstBlockUsed_ = 0;
    firstBlockSpare_ = freeBlocks_->size;
  }
  char *tem = freeBlocks_->mem + firstBlockUsed_;
  firstBlockUsed_ += n;
  firstBlockSpare_ -= n;
  return tem;
}

void AppEventHandler::message(MessageEvent *event)
{
  if (event->message().isError()) {
    SGMLApplication::ErrorEvent appEvent;
    switch (event->message().type->severity) {
    case MessageType::quantityError:
      appEvent.type = SGMLApplication::ErrorEvent::quantity;
      break;
    default:
      appEvent.type = SGMLApplication::ErrorEvent::other;
      break;
    }
    setLocation(appEvent.pos, event->message().loc);
    app_->error(appEvent);
    nErrors_++;
  }
  if (!*messagesInhibitedPtr_)
    reporter_->reportMessage(event->message());
  delete event;
}

void AppEventHandler::startElement(StartElementEvent *event)
{
  SGMLApplication::StartElementEvent appEvent;
  setString(appEvent.gi, event->name());
  const ElementDefinition *def = event->elementType()->definition();
  switch (def->declaredContent()) {
  case ElementDefinition::modelGroup:
    appEvent.contentType
      = (def->compiledModelGroup()->containsPcdata()
	 ? SGMLApplication::StartElementEvent::mixed
	 : SGMLApplication::StartElementEvent::element);
    break;
  case ElementDefinition::any:
    appEvent.contentType = SGMLApplication::StartElementEvent::mixed;
    break;
  case ElementDefinition::cdata:
    appEvent.contentType = SGMLApplication::StartElementEvent::cdata;
    break;
  case ElementDefinition::rcdata:
    appEvent.contentType = SGMLApplication::StartElementEvent::rcdata;
    break;
  case ElementDefinition::empty:
    appEvent.contentType = SGMLApplication::StartElementEvent::empty;
    break;
  }
  appEvent.included = event->included();
  appEvent.nAttributes = event->attributes().length();
  if (appEvent.nAttributes != 0) {
    if (event->attributes().conref())
      appEvent.contentType = SGMLApplication::StartElementEvent::empty;
    setAttributes(appEvent.attributes, event->attributes());
  }
  setLocation(appEvent.pos, event->location());
  app_->startElement(appEvent);
  freeAll();
  delete event;
}

void AppEventHandler::endElement(EndElementEvent *event)
{
  SGMLApplication::EndElementEvent appEvent;
  setString(appEvent.gi, event->name());
  setLocation(appEvent.pos, event->location());
  app_->endElement(appEvent);
  delete event;
}

void AppEventHandler::data(DataEvent *event)
{
  SGMLApplication::DataEvent appEvent;
  appEvent.data.ptr = event->data();
  appEvent.data.len = event->dataLength();
  setLocation(appEvent.pos, event->location());
  app_->data(appEvent);
  delete event;
}

void AppEventHandler::pi(PiEvent *event)
{
  SGMLApplication::PiEvent appEvent;
  appEvent.data.ptr = event->data();
  appEvent.data.len = event->dataLength();
  const Entity *entity = event->entity();
  if (entity)
    setString(appEvent.entityName, entity->name());
  else
    appEvent.entityName.len = 0;
  setLocation(appEvent.pos, event->location());
  app_->pi(appEvent);
  delete event;
}

void AppEventHandler::sdataEntity(SdataEntityEvent *event)
{
  SGMLApplication::SdataEvent appEvent;
  appEvent.text.ptr = event->data();
  appEvent.text.len = event->dataLength();
  setString(appEvent.entityName, event->entity()->name());
  // Don't want location of chars in entity.
  setLocation(appEvent.pos, event->location().origin()->parent());
  app_->sdata(appEvent);
  delete event;
}

void AppEventHandler::externalDataEntity(ExternalDataEntityEvent *event)
{
  SGMLApplication::ExternalDataEntityRefEvent appEvent;
  setEntity(appEvent.entity, *event->entity());
  setLocation(appEvent.pos, event->location());
  app_->externalDataEntityRef(appEvent);
  freeAll();
  delete event;
}

void AppEventHandler::subdocEntity(SubdocEntityEvent *event)
{
  SGMLApplication::SubdocEntityRefEvent appEvent;
  setEntity(appEvent.entity, *event->entity());
  setLocation(appEvent.pos, event->location());
  app_->subdocEntityRef(appEvent);
  freeAll();
  delete event;
}

void AppEventHandler::startDtd(StartDtdEvent *event)
{
  SGMLApplication::StartDtdEvent appEvent;
  setString(appEvent.name, event->name());
  const Entity *entity = event->entity().pointer();
  if (entity) {
    appEvent.haveExternalId = 1;
    setExternalId(appEvent.externalId,
		  entity->asExternalEntity()->externalId());
  }
  else
    appEvent.haveExternalId = 0;
  setLocation(appEvent.pos, event->location());
  app_->startDtd(appEvent);
  freeAll();
  delete event;
}

void AppEventHandler::endDtd(EndDtdEvent *event)
{
  SGMLApplication::EndDtdEvent appEvent;
  setString(appEvent.name, event->dtd().name());
  setLocation(appEvent.pos, event->location());
  app_->endDtd(appEvent);
  delete event;
}

void AppEventHandler::endProlog(EndPrologEvent *event)
{
  SGMLApplication::EndPrologEvent appEvent;
  setLocation(appEvent.pos, event->location());
  app_->endProlog(appEvent);
  delete event;
}

void AppEventHandler::appinfo(AppinfoEvent *event)
{
  SGMLApplication::AppinfoEvent appEvent;
  const CString *str;
  if (event->literal(str)) {
    setString(appEvent.string, *str);
    appEvent.none = 0;
  }
  else
    appEvent.none = 1;
  setLocation(appEvent.pos, event->location());
  app_->appinfo(appEvent);
  delete event;
}

void AppEventHandler::commentDecl(CommentDeclEvent *event)
{
  SGMLApplication::CommentDeclEvent appEvent;
  appEvent.nComments = 0;
  size_t i;
  for (i = 0; i < event->decl().length(); i++)
    if (event->decl()[i].type == DeclItem::comment)
      appEvent.nComments++;
  SGMLApplication::CharString *comments
    = (SGMLApplication::CharString *)allocate(appEvent.nComments * 2
					      * sizeof(SGMLApplication::CharString));
  appEvent.comments = comments;
  appEvent.seps = appEvent.comments + appEvent.nComments;
  size_t j = 0;
  for (i = 0; i < event->decl().length(); i++)
    switch (event->decl()[i].type) {
    case DeclItem::comment:
      setString(comments[j++], event->decl()[i].str);
      clearString(comments[appEvent.nComments + j - 1]);
      break;
    case DeclItem::s:
      setString(comments[appEvent.nComments + j - 1], event->decl()[i].str);
      break;
    default:
      break;
    }
  setLocation(appEvent.pos, event->location());
  app_->commentDecl(appEvent);
  freeAll();
  delete event;
}

void AppEventHandler::markedSectionStart(MarkedSectionStartEvent *event)
{
  SGMLApplication::MarkedSectionStartEvent appEvent;
  size_t i;
  const Decl &decl = event->decl();
  unsigned depth = 0;
  appEvent.nParams = 0;
  for (i = 0; i < decl.length(); i++)
    switch (decl[i].type) {
    case DeclItem::reservedName:
      if (!depth)
	appEvent.nParams++;
      break;
    case DeclItem::entityStart:
      if (!depth)
	appEvent.nParams++;
      depth++;
      break;
    case DeclItem::entityEnd:
      depth--;
      break;
    default:
      break;
    }
  SGMLApplication::MarkedSectionStartEvent::Param *params
    = (SGMLApplication::MarkedSectionStartEvent::Param *)
      allocate(appEvent.nParams * sizeof(appEvent.params[0]));
  appEvent.params = params;
  size_t j = 0;
  for (i = 0; i < decl.length(); i++)
    switch (decl[i].type) {
    case DeclItem::reservedName:
      if (!depth) {
	switch (decl[i].reservedNameIndex) {
	case Syntax::rTEMP:
	  params[j].type
	    = SGMLApplication::MarkedSectionStartEvent::Param::temp;
	  break;
	case Syntax::rINCLUDE:
	  params[j].type
	    = SGMLApplication::MarkedSectionStartEvent::Param::include;
	  break;
	case Syntax::rRCDATA:
	  params[j].type
	    = SGMLApplication::MarkedSectionStartEvent::Param::rcdata;
	  break;
	case Syntax::rCDATA:
	  params[j].type
	    = SGMLApplication::MarkedSectionStartEvent::Param::cdata;
	  break;
	case Syntax::rIGNORE:
	  params[j].type
	    = SGMLApplication::MarkedSectionStartEvent::Param::ignore;
	  break;
	default:
	  CANNOT_HAPPEN();
	}
	clearString(params[j].entityName);
	j++;
      }
      break;
    case DeclItem::entityStart:
      if (!depth) {
	params[j].type
	  = SGMLApplication::MarkedSectionStartEvent::Param::entityRef;
	setString(params[j].entityName,
		  decl[i].loc.origin()->asEntityOrigin()->entity()->name());
	j++;
      }
      depth++;
      break;
    case DeclItem::entityEnd:
      depth--;
      break;
    default:
      break;
    }
  switch (event->status()) {
  case MarkedSectionEvent::include:
    appEvent.status = SGMLApplication::MarkedSectionEvent::include;
    break;
  case MarkedSectionEvent::rcdata:
    appEvent.status = SGMLApplication::MarkedSectionEvent::rcdata;
    break;
  case MarkedSectionEvent::cdata:
    appEvent.status = SGMLApplication::MarkedSectionEvent::cdata;
    break;
  case MarkedSectionEvent::ignore:
    appEvent.status = SGMLApplication::MarkedSectionEvent::ignore;
    break;
  }
  setLocation(appEvent.pos, event->location());
  app_->markedSectionStart(appEvent);
  freeAll();
  delete event;
}

void AppEventHandler::ignoredChars(IgnoredCharsEvent *event)
{
  SGMLApplication::IgnoredCharsEvent appEvent;
  appEvent.data.ptr = event->data();
  appEvent.data.len = event->dataLength();
  setLocation(appEvent.pos, event->location());
  app_->ignoredChars(appEvent);
  delete event;
}

void AppEventHandler::markedSectionEnd(MarkedSectionEndEvent *event)
{
  SGMLApplication::MarkedSectionEndEvent appEvent;
  switch (event->status()) {
  case MarkedSectionEvent::include:
    appEvent.status = SGMLApplication::MarkedSectionEvent::include;
    break;
  case MarkedSectionEvent::rcdata:
    appEvent.status = SGMLApplication::MarkedSectionEvent::rcdata;
    break;
  case MarkedSectionEvent::cdata:
    appEvent.status = SGMLApplication::MarkedSectionEvent::cdata;
    break;
  case MarkedSectionEvent::ignore:
    appEvent.status = SGMLApplication::MarkedSectionEvent::ignore;
    break;
  }
  setLocation(appEvent.pos, event->location());
  app_->markedSectionEnd(appEvent);
  delete event;
}

void AppEventHandler::setLocation1(SGMLApplication::Position &pos,
				   const Location &loc)
{
  const Location *locp = &loc;
  for (;;) {
    if (locp->origin().isNull()) {
      lastOrigin_.clear();
      openEntityPtr_ = (SpOpenEntity *)0;
      return;
    }
    const EntityOrigin *origin = locp->origin()->asEntityOrigin();
    if (origin && origin->externalInfo())
      break;
    locp = &locp->origin()->parent();
  }
  lastOrigin_ = locp->origin();
  pos = locp->index();
  openEntityPtr_ = new SpOpenEntity(locp->origin());
  app_->openEntityChange(openEntityPtr_);
}

void 
AppEventHandler::setAttributes(const SGMLApplication::Attribute *&attributes,
			       const AttributeList &attributeList)
{
  size_t nAttributes = attributeList.length();
  SGMLApplication::Attribute *to
    = (SGMLApplication::Attribute *)allocate(nAttributes * sizeof(*to));
  attributes = to;
  for (size_t i = 0; i < nAttributes; i++) {
    SGMLApplication::Attribute *p = to + i;
    setString(p->name, attributeList.name(i));
    const AttributeValue *value = attributeList.value(i);
    if (!value)
      p->type = SGMLApplication::Attribute::invalid;
    else {
      const Text *text;
      const CString *string;
      switch (value->info(text, string)) {
      case AttributeValue::implied:
	p->type = SGMLApplication::Attribute::implied;
	break;
      case AttributeValue::tokenized:
	{
	  if (attributeList.specified(i))
	    p->defaulted = SGMLApplication::Attribute::specified;
	  else if (attributeList.current(i))
	    p->defaulted = SGMLApplication::Attribute::current;
	  else
	    p->defaulted = SGMLApplication::Attribute::definition;
	  p->type = SGMLApplication::Attribute::tokenized;
	  p->nEntities = 0;
	  p->notation.name.len = 0;
	  p->isId = attributeList.id(i);
	  setString(p->tokens, *string);
	  const AttributeSemantics *semantics = attributeList.semantics(i);
	  if (semantics) {
	    ConstResourcePointer<Notation> notation = semantics->notation();
	    if (!notation.isNull())
	      setNotation(p->notation, *notation);
	    else {
	      size_t nEntities = semantics->nEntities();
	      if (nEntities) {
		SGMLApplication::Entity *v
		  = (SGMLApplication::Entity *)allocate(nEntities * sizeof(*v));
		p->entities = v;
		p->nEntities = nEntities;
		for (size_t i = 0; i < nEntities; i++)
		  setEntity(v[i], *semantics->entity(i));
	      }
	    }
	  }
	}
	break;
      case AttributeValue::cdata:
	{
	  p->type = SGMLApplication::Attribute::cdata;
	  if (attributeList.specified(i))
	    p->defaulted = SGMLApplication::Attribute::specified;
	  else if (attributeList.current(i))
	    p->defaulted = SGMLApplication::Attribute::current;
	  else
	    p->defaulted = SGMLApplication::Attribute::definition;
	  EsisTextIter::Type type;
	  const Char *s;
	  size_t length;
	  const Entity *entity;
	  size_t nChunks = 0;
	  {
	    EsisTextIter iter(*text);
	    while (iter.next(type, s, length, entity))
	      nChunks++;
	  }
	  p->cdataChunks
	    = (SGMLApplication::Attribute::CdataChunk *)allocate(nChunks * sizeof(SGMLApplication::Attribute::CdataChunk));
	  p->nCdataChunks = nChunks;

	  {
	    size_t i = 0;
	    for (EsisTextIter iter(*text);
		 iter.next(type, s, length, entity);
		 i++) {
	      SGMLApplication::Attribute::CdataChunk *chunk
		= (SGMLApplication::Attribute::CdataChunk *)(p->cdataChunks + i);
	      if (type == EsisTextIter::data)
		chunk->isSdata = 0;
	      else {
		chunk->isSdata = 1;
		setString(chunk->entityName, entity->name());
	      }
	      chunk->data.ptr = s;
	      chunk->data.len = length;
	    }
	  }
	}
	break;
      }
    }
  }
}


void AppEventHandler::setEntity(SGMLApplication::Entity &to,
				const Entity &from)
{
  setString(to.name, from.name());
  switch (from.declType()) {
  case Entity::generalEntity:
    to.declType = SGMLApplication::Entity::general;
    break;
  case Entity::parameterEntity:
    to.declType = SGMLApplication::Entity::parameter;
    break;
  case Entity::doctype:
    to.declType = SGMLApplication::Entity::doctype;
    break;
  case Entity::linktype:
    to.declType = SGMLApplication::Entity::linktype;
    break;
  }
  switch (from.dataType()) {
  case Entity::sgmlText:
    to.dataType = SGMLApplication::Entity::sgml;
    break;
  case Entity::cdata:
    to.dataType = SGMLApplication::Entity::cdata;
    break;
  case Entity::sdata:
    to.dataType = SGMLApplication::Entity::sdata;
    break;
  case Entity::ndata:
    to.dataType = SGMLApplication::Entity::ndata;
    break;
  case Entity::subdoc:
    to.dataType = SGMLApplication::Entity::subdoc;
    break;
  case Entity::pi:
    to.dataType = SGMLApplication::Entity::pi;
    break;
  }
  const InternalEntity *internal = from.asInternalEntity();
  if (internal) {
    to.isInternal = 1;
    setString(to.text, internal->string());
  }
  else {
    const ExternalEntity *external = from.asExternalEntity();
    to.isInternal = 0;
    setExternalId(to.externalId, external->externalId());
    const ExternalDataEntity *externalData = from.asExternalDataEntity();
    if (externalData) {
      setNotation(to.notation, *externalData->notation());
      to.nAttributes = externalData->attributes().length();
      if (to.nAttributes)
	setAttributes(to.attributes, externalData->attributes());
    }
    else {
      to.notation.name.len = 0;
      to.nAttributes = 0;
    }
  }
}


void AppEventHandler::setNotation(SGMLApplication::Notation &to,
				   const Notation &from)
{
  setString(to.name, from.name());
  setExternalId(to.externalId, from.externalId());
}

void AppEventHandler::setExternalId(SGMLApplication::ExternalId &to,
				    const ExternalId &from)
{
  const CString *str;
  str = from.systemIdString();
  if (str) {
    to.haveSystemId = 1;
    setString(to.systemId, *str);
  }
  else
    to.haveSystemId = 0;
  str = from.publicIdString();
  if (str) {
    to.havePublicId = 1;
    setString(to.publicId, *str);
  }
  else
    to.havePublicId = 0;
}

SpOpenEntity::SpOpenEntity(const ConstResourcePointer<Origin> &origin)
: origin_(origin)
{
}

SGMLApplication::Location
SpOpenEntity::location(SGMLApplication::Position pos) const
{
  SGMLApplication::Location loc;
  const Origin *origin = origin_.pointer();
  const EntityOrigin *entityOrigin;
  const ExternalInfo *externalInfo;
  Index index = Index(pos);
  for (;;) {
    if (!origin)
      return loc;
    entityOrigin = origin->asEntityOrigin();
    if (entityOrigin) {
      externalInfo = entityOrigin->externalInfo();
      if (externalInfo)
	break;
    }
    const Location &loc = origin->parent();
    index = loc.index();
    origin = loc.origin().pointer();
  }
  NamedCharRef ref;
  Offset off;
  ConstResourcePointer<Entity> entity(entityOrigin->entity());
  if (!entity.isNull())
    AppEventHandler::setString(loc.entityName, entity->name());
  while (entityOrigin->lookupIndex(index, off, ref)
	 == EntityOrigin::replacementCharIndex)
    index = ref.refStartIndex();
  loc.entityOffset = off;
  StorageObjectLocation soLoc;
  if (!EntityManager::externalize(externalInfo, off, soLoc))
    return loc;
  loc.lineNumber = soLoc.lineNumber;
  AppEventHandler::setString(loc.filename, soLoc.storageObjectSpec->id);
  loc.columnNumber = soLoc.columnNumber;
  loc.byteOffset = soLoc.byteIndex;
  loc.other = soLoc.storageObjectSpec;
  return loc;
}
