// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif
#include "config.h"
#include "EasyParser.H"
#include "PosixStorage.H"
#include "PosixHttpStorage.H"
#include "macros.H"
#ifdef MULTI_BYTE
#include "UTF8CodingSystem.H"
#include "UCS2CodingSystem.H"
#include "UnicodeCodingSystem.H"
#include "UJISCodingSystem.H"
#include "SJISCodingSystem.H"
#endif
#include "IdentityCodingSystem.H"
#include "Vector.H"
#include "EntityManager.H"

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_GETTEXT
extern "C" {
extern char *dgettext(const char *, const char *);
extern char *gettext(const char *);
extern char *textdomain(const char *);
extern char *bindtextdomain(const char *, const char *);
}
#endif /* HAVE_GETTEXT */

#ifdef MULTI_BYTE
static UnivCharsetDesc::Range range = { 0, 65536, 0 };
#else
static UnivCharsetDesc::Range range = { 0, 256, 0 };
#endif

#ifdef MULTI_BYTE
static UTF8CodingSystem utf8CodingSystem;
static UCS2CodingSystem ucs2CodingSystem;
static UnicodeCodingSystem unicodeCodingSystem;
static UJISCodingSystem ujisCodingSystem;
static SJISCodingSystem sjisCodingSystem;
#endif
static IdentityCodingSystem identityCodingSystem;

static struct {
  const char *name;
  const CodingSystem *cs;
} codingSystems[] = {
#ifdef MULTI_BYTE
  { "UTF8", &utf8CodingSystem },
  { "UCS2", &ucs2CodingSystem },
  { "UNICODE", &unicodeCodingSystem },
  { "UJIS", &ujisCodingSystem },
  { "SJIS", &sjisCodingSystem },
#endif
  { "IDENTITY", &identityCodingSystem },
};

EasyParser::EasyParser(const char *codingName)
: storageManager_(0),
  fdStorageManager_(0),
  systemCharset_(&range, 1),
  MessageReporter(&errorStream_)
{
  codingSystem_ = 0;
  if (codingName && strlen(codingName) < 50) {
    char buf[50];
    strcpy(buf, codingName);
    for (char *p = buf; *p; p++)
      *p = toupper((unsigned char)*p);
    for (int i = 0; i < SIZEOF(codingSystems); i++)
      if (strcmp(buf, codingSystems[i].name) == 0) {
	codingSystem_ = codingSystems[i].cs;
	break;
      }
  }
  if (!codingSystem_)
    codingSystem_ = &identityCodingSystem;
  errorStream_.open(cerr.rdbuf(), codingSystem_);
#ifndef MESSAGE_DOMAIN
#define MESSAGE_DOMAIN "sp"
#endif
#ifdef HAVE_GETTEXT
  const char *dir = getenv("TEXTDOMAINDIR");
  if (dir)
    bindtextdomain(MESSAGE_DOMAIN, dir);
#endif
}

EasyParser::~EasyParser()
{
  delete storageManager_;
  delete fdStorageManager_;
}

void EasyParser::setErrorStream(streambuf *sbuf)
{
  errorStream_.open(sbuf, codingSystem_);
}

#ifdef __GNUG__
typedef Vector<CString> Dummy_Vector_CString;
#endif

void EasyParser::init(int nFiles, char *const *files,
		      unsigned flags)
{
  storageManager_ = new PosixStorageManager("FILE",
					    systemCharset_,
					    codingSystem_,
					    5);
  entityManager_.init(storageManager_, codingSystem_, systemCharset_);
  fdStorageManager_ = new PosixFdStorageManager("FD", systemCharset_);
  entityManager_.registerStorageManager(fdStorageManager_);
  entityManager_.registerStorageManager(new PosixHttpStorageManager("HTTP"));
  int i;
  for (i = 0; i < SIZEOF(codingSystems); i++)
    entityManager_.registerCodingSystem(codingSystems[i].name,
					codingSystems[i].cs);
  parser_.init(entityManager_.externalEntityOpener(),
	       openDocumentEntity(nFiles, files),
	       systemCharset_,
	       options);
  for (i = 0; i < catalogSysids_.length(); i++)
    // filenames specified on command-line must exist
    entityManager_.addCatalogSysid(codingSystem_->convertIn(catalogSysids_[i]), 1);
  const char *e = getenv("SGML_CATALOG_FILES");
  if (!e)
    e = "CATALOG";
  // FIXME Should probably convert before splitting.
  String<char> str(e, strlen(e) + 1);
  char *start = str.pointer();
  char *p = start;
  for (;;) {
    if (*p == '\0') {
      entityManager_.addCatalogSysid(codingSystem_->convertIn(start));
      break;
    }
#ifdef MSDOS_FILENAMES
#define FILE_SEP ';'
#else
#define FILE_SEP ':'
#endif
    if (*p == FILE_SEP) {
      *p++ = '\0';
      entityManager_.addCatalogSysid(codingSystem_->convertIn(start));
      start = p;
    }
    else
      p++;
  }
  for (i = 0; i < includes_.length(); i++)
    parser_.addInclude(codingSystem_->convertIn(includes_[i]));
  for (i = 0; i < activeLinkTypes_.length(); i++) 
    parser_.activateLinkType(codingSystem_->convertIn(activeLinkTypes_[i]));
  if (flags & allLinkTypesActivated)
    parser_.allLinkTypesActivated();
}

InputSource *EasyParser::openDocumentEntity(int nFiles,
					    char *const *files)
{
  Vector<CString> filenames(nFiles == 0 ? 1 : nFiles);
  int i;
  for (i = 0; i < nFiles; i++)
    filenames[i] = codingSystem_->convertIn(strcmp(files[i], "-") == 0
					    ? "FD:0"
					    : files[i]);
  if (nFiles == 0)
    filenames[0] = codingSystem_->convertIn("FD:0");
  return entityManager_.openDocumentEntity(filenames.pointer(),
					   filenames.length(),
					   systemCharset_,
					   *this);
}


CString EasyParser::messageText(const char *domain, const char *s)
{
#ifdef HAVE_GETTEXT
  s = dgettext(MESSAGE_DOMAIN, s);
#endif
  return codingSystem_->convertIn(s);
}

void EasyParser::addCatalogSysid(const char *s)
{
  catalogSysids_.grow() = s;
}

void EasyParser::addInclude(const char *s)
{
  includes_.grow() = s;
}

void EasyParser::activateLinkType(const char *s)
{
  activeLinkTypes_.grow() = s;
}
