// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#include "config.h"
#include "MessageReporter.H"
#include "EntityManager.H"
#include "StorageManager.H"
#include "macros.H"
#include "rtti.H"
#include "MessageArg.H"
#include "ErrnoMessageArg.H"
#include "Entity.H"

#include <string.h>
extern "C" {
  char *strerror(int);
}

static const char systemDomain[] = "system";

const char MessageReporter::domain_[] = "MessageReporter";

MessageReporter::MessageReporter(OutputCharStream *errsp)
: os_(errsp), options_(0)
{
}

void MessageReporter::addOption(Option option)
{
  options_ |= option;
}

void MessageReporter::dispatchMessage(Message &msg)
{
  reportMessage(msg);
}

void MessageReporter::initMessage(Message &msg)
{
  msg.loc = currentLocation_;
}

void MessageReporter::reportMessage(const Message &message)
{
  Offset off;
  const ExternalInfo *externalInfo = locationHeader(message.loc, off);
  if (externalInfo) {
    printLocation(externalInfo, off);
    os() << ':';
  }
  const char *msg;
  switch (message.type->severity) {
  case MessageType::info:
    msg = "I";
    break;
  case MessageType::warning:
    msg = "W";
    break;
  case MessageType::quantityError:
    msg = "Q";
    break;
  case MessageType::error:
    msg = "E";
    break;
  default:
    CANNOT_HAPPEN();
  }
  os() << messageText(domain_, msg) << ": ";
  for (int id = 0; id < 2; id++) {
    if (id == 1) {
      if (!message.type->hasAuxLoc())
	break;
      Offset off;
      const ExternalInfo *externalInfo = locationHeader(message.auxLoc, off);
      printLocation(externalInfo, off);
      os() << ": ";
    }
    CString text = messageText(message.type->domain,
			       id == 0
			       ? message.type->id
			       : message.type->auxId);
    size_t i = 0;
    while (i < text.length()) {
      if (text[i] == '%') {
	i++;
	if (i >= text.length())
	  break;
	if (text[i] >= '1' && text[i] <= '9') {
	  if (text[i] - '1' < message.args.length())
	    message.args[text[i] - '1']->append(*this);
	}
	else
	  os().put(text[i]);
	i++;
      }
      else {
	os().put(text[i]);
	i++;
      }
    }
    os() << '\n';
  }
  if ((options_ & openElements) && message.openElementInfo.length() > 0) {
    if (externalInfo) {
      printLocation(externalInfo, off);
      os() << ": ";
    }
    os() << messageText(domain_, "open elements");
    os() << ':';
    unsigned nOpenElements = message.openElementInfo.length();
    for (unsigned i = 0;; i++) {
      if (i > 0
	  && (i == nOpenElements || message.openElementInfo[i].included)) {
	// describe last match in previous open element
	const OpenElementInfo &prevInfo = message.openElementInfo[i - 1];
	if (prevInfo.matchType.length() != 0) {
	  os() << " (" << prevInfo.matchType;
	  if (prevInfo.matchIndex != 0)
	    os() << '[' << (unsigned long)prevInfo.matchIndex << ']';
	  os() << ')';
	}
      }
      if (i == nOpenElements)
	break;
      const OpenElementInfo &e = message.openElementInfo[i];
      os() << ' ' << e.gi;
      if (i > 0 && !e.included) {
	unsigned long n = message.openElementInfo[i - 1].matchIndex;
	if (n != 0)
	  os() << '[' << n << ']';
      }
    }
    os() << '\n';
  }
  os().flush();
}

const ExternalInfo *MessageReporter::locationHeader(const Location &loc,
						    Offset &off)
{
  const Origin *origin = loc.origin().pointer();
  Index index = loc.index();
  if (!(options_ & openEntities)) {
    while (origin) {
      const EntityOrigin *entityOrigin = origin->asEntityOrigin();
      if (entityOrigin) {
	const ExternalInfo *externalInfo = entityOrigin->externalInfo();
	if (externalInfo) {
	  NamedCharRef ref;
	  while (entityOrigin->lookupIndex(index, off, ref)
		 == EntityOrigin::replacementCharIndex)
	    index = ref.refStartIndex();
	  return externalInfo;
	}
      }
      const Location &loc = origin->parent();
      index = loc.index() + origin->refLength();
      origin = loc.origin().pointer();
    }
  }
  else {
    Boolean doneHeader = 0;
    while (origin) {
      const EntityOrigin *entityOrigin = origin->asEntityOrigin();
      if (entityOrigin) {
	if (!doneHeader) {
	  if (!entityOrigin->entity().isNull()) {
	    Offset parentOff;
	    Location parentLoc = entityOrigin->parent();
	    parentLoc += entityOrigin->refLength();
	    const ExternalInfo *parentInfo = locationHeader(parentLoc, parentOff);
	    if (parentInfo) {
	      CString text = messageText(domain_,
					 "In entity %1 included from %2");
	      for (size_t i = 0; i < text.length(); i++) {
		if (text[i] == '%') {
		  if (i + 1 < text.length()) {
		    i++;
		    if (text[i] == '1')
		      os() << entityOrigin->entity()->name();
		    else if (text[i] == '2')
		      printLocation(parentInfo, parentOff);
		    else if (text[i] >= '3' && text[i] <= '9')
		      ;
		    else
		      os().put(text[i]);
		  }
		}
		else
		  os().put(text[i]);
	      }
	      os() << '\n';
	    }
	  }
	  doneHeader = 1;
	}
	NamedCharRef ref;
	while (entityOrigin->lookupIndex(index, off, ref)
	       == EntityOrigin::replacementCharIndex)
	  index = ref.refStartIndex();
	const ExternalInfo *externalInfo = entityOrigin->externalInfo();
	if (externalInfo)
	  return externalInfo;
	const InternalEntity *internal
	  = entityOrigin->entity()->asInternalEntity();
	if (!internal)
	  break;
	Location loc = internal->text().charLocation(off);
	index = loc.index();
	origin = loc.origin().pointer();
      }
      else {
	const Location &loc = origin->parent();
	index = loc.index() + origin->refLength();
	origin = loc.origin().pointer();
      }
    }
  }
  return 0;
}

void MessageReporter::printLocation(const ExternalInfo *externalInfo,
				    Offset off)
{
  if (!externalInfo) {
    os() << messageText(domain_, "(invalid location)");
    return;
  }
  StorageObjectLocation soLoc;
  if (!EntityManager::externalize(externalInfo, off, soLoc)) {
    os() << messageText(domain_, "(invalid location)");
    return;
  }
  if (strcmp(soLoc.storageObjectSpec->storageManager->type(), "FILE") != 0)
    os() << soLoc.storageObjectSpec->storageManager->type() << ':';
  os() << soLoc.storageObjectSpec->id;
  if (soLoc.lineNumber == (unsigned long)-1)
    os() << ": " << messageText(domain_, "offset ")
      << soLoc.storageObjectOffset;
  else
    os() << ':' << soLoc.lineNumber << ':' << soLoc.columnNumber;
#if 0
  if (soLoc.byteIndex != (unsigned long)-1)
    os() << ':' << soLoc.byteIndex;
#endif
}

void MessageReporter::appendNumber(unsigned long n)
{
  os() << n;
}

void MessageReporter::appendOrdinal(unsigned long n)
{
  os() << n;
  switch (n % 10) {
  case 1:
    os() << messageText(domain_, "st");
    break;
  case 2:
    os() << messageText(domain_, "nd");
    break;
  case 3:
    os() << messageText(domain_, "rd");
    break;
  default:
    os() << messageText(domain_, "th");
    break;
  }
}

void MessageReporter::appendChars(const Char *p, size_t n)
{
  os().put('`').write(p, n).put('\'');
}

void MessageReporter::appendOther(const OtherMessageArg *p)
{
  const ErrnoMessageArg *ea = ptr_cast(ErrnoMessageArg, p);
  
  if (ea) {
    os() << messageText(systemDomain, strerror(ea->errnum()));
    return;
  }
  os() << messageText(domain_, "(invalid argument type)");
}

void MessageReporter::appendText(const char *domain, const char *id)
{
  os() << messageText(domain, id);
}
