// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifndef Catalog_INCLUDED
#define Catalog_INCLUDED 1

#ifdef __GNUG__
#pragma interface
#endif

#include "Message.H"
#include "CString.H"
#include "types.H"
#include "GrowableVectorD.H"
#include "HashTable.H"
#include "InputSource.H"
#include "Boolean.H"
#include "SubstTable.H"

class CharsetInfo;
class MessageArg;
class PublicId;

struct CatalogEntry {
  CString to;
  Location loc;
  size_t catalogNumber;
};

class Catalog {
public:
  enum DeclType { entityDecl, doctypeDecl, linktypeDecl };
  enum { nDeclType = 3 };
  Catalog();
  Boolean lookupEntity(const PublicId *,
		       const CString &name, DeclType,
		       const SubstTable<Char> *substTable,
		       const CString *&systemId,
		       Location &specLoc) const;
  Boolean sgmlDecl(const CString *&systemId, Location &specLoc) const;
  Boolean override() const;
  void addPublicId(CString &publicID, CString &systemId, const Location &);
  void addName(CString &name, DeclType, CString &systemId, const Location &);
  void setSgmlDecl(CString &str, const Location &loc);
  void setOverride();
  void endCatalog();
private:
  Catalog(const Catalog &);	// undefined
  void operator=(const Catalog &); // undefined
  HashTable<CString,CatalogEntry> publicIds_;
  HashTable<CString,CatalogEntry> names_[nDeclType];
  size_t catalogNumber_;
  Boolean override_;
  Boolean haveSgmlDecl_;
  CString sgmlDecl_;
  Location sgmlDeclLoc_;
};

class CatalogParser : private Messenger {
public:
  CatalogParser(const CharsetInfo &);
  void parseCatalog(Catalog *, InputSource *, Messenger &mgr);
public:
  // Since it's a return type, it has to be public to keep some 
  // (broken) compilers happy.
  enum Param {
    eofParam,
    literalParam,
    nameParam,
    percentParam
  };
private:
  enum {
    data,
    eof,
    nul,
    lit,
    minus,
    s,
    min				// other minimum data characters
  };
  enum { minimumLiteral = 01 };

  Messenger &messenger() { return *this; }
  void dispatchMessage(Message &);
  void initMessage(Message &);
  void parsePublic();
  void parseNameMap(Catalog::DeclType declType);
  Param parseParam(unsigned flags = 0);
  Boolean parseArg();
  void parseLiteral(Char delim, unsigned flags);
  void parseName();
  void skipComment();
  void upcase(CString &);
  Boolean isMinimumData(Xchar c) {
    int cat = categoryTable_[c];
    return (cat == min || (cat == s && c != tab_) || cat == minus);
  }
  Xchar get() { return in_->get(messenger()); }
  void unget() { in_->ungetToken(); }
  Messenger *mgr_;
  InputSource *in_;
  Catalog *catalog_;
  CString param_;
  Location paramLoc_;
  Char minus_;
  Char tab_;
  Char rs_;
  Char re_;
  Char space_;
  CString publicKey_;
  CString entityKey_;
  CString doctypeKey_;
  CString linktypeKey_;
  CString overrideKey_;
  CString sgmlDeclKey_;
  XcharMap<unsigned char> categoryTable_;
  SubstTable<Char> substTable_;
};

inline
Boolean Catalog::override() const
{
  return override_;
}

inline
void Catalog::setOverride()
{
  override_ = 1;
}

#endif /* not Catalog_INCLUDED */
