// Copyright (c) 1995 James Clark
// See the file COPYING for copying permission.

#include "config.h"
#include "PosixHttpStorage.H"
#include "PosixHttpStorageMessages.H"
#include "PosixBaseStorageObject.H"
#include "UnivCharsetDesc.H"
#include "MessageArg.H"
#include "ErrnoMessageArg.H"
#include "macros.H"
#include <sys/types.h>

#ifdef HAVE_SOCKET
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif /* HAVE_SOCKET */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <ctype.h>

#ifdef INCLUDE_IO_H
#include <io.h>		// for open, fstat, lseek, read prototypes
#endif

#ifdef INCLUDE_UNISTD_H
#include <unistd.h>
#endif

#ifdef INCLUDE_OSFCN_H
#include <osfcn.h>
#endif

static UnivCharsetDesc::Range range = { 0, 256, 0 };
static UnivCharsetDesc latin1Charset(&range, 1);

class PosixHttpStorageObject : public PosixBaseStorageObject {
public:
  PosixHttpStorageObject(String<char> &host, unsigned short port, String<char> &path);
  ~PosixHttpStorageObject();
  Boolean open(Messenger &, Boolean mayRewind, size_t &blockSize);
  Boolean read(char *buf, size_t bufSize, Messenger &mgr, size_t &nread);
  Boolean seek(off_t, Messenger &);
private:
  PosixHttpStorageObject(const PosixHttpStorageObject &); // undefined
  void operator=(const PosixHttpStorageObject &); // undefined
  Boolean readHeader(Messenger &);
  Boolean readLine(Messenger &mgr, String<char> &line, String<char> &leftOver);
  unsigned short port_;
  String<char> host_;
  CString hostStr_;
  String<char> path_;
};

PosixHttpStorageManager::PosixHttpStorageManager(const char *type)
: type_(type), StorageManager(latin1Charset)
{
}

const char *PosixHttpStorageManager::type() const
{
  return type_;
}

Boolean PosixHttpStorageManager::inheritable() const
{
  return 1;
}

StorageObject *PosixHttpStorageManager::makeStorageObject(const CString &id,
							  Messenger &mgr)
{
  if (id.length() < 2 || id[0] != '/' || id[1] != '/') {
    mgr.message(PosixHttpStorageMessages::badRelative,
		StringMessageArg(id));
    return 0;
  }
  size_t i = 2;
  String<char> host;
  while (i < id.length()) {
    if (id[i] == '/')
      break;
    if (id[i] == ':')
      break;
    host += char(id[i]);
    i++;
  }
  if (host.length() == 0) {
    mgr.message(PosixHttpStorageMessages::emptyHost,
		StringMessageArg(id));
    return 0;
  }
  unsigned short port;
  if (i < id.length() && id[i] == ':') {
    i++;
    String<char> digits;
    while (i < id.length() && id[i] != '/') {
      digits += char(id[i]);
      i++;
    }
    if (digits.length() == 0) {
      mgr.message(PosixHttpStorageMessages::emptyPort,
		  StringMessageArg(id));
      return 0;
    }
    digits += '\0';
    char *endptr;
    long n = strtol(digits.pointer(), &endptr, 10);
    if (endptr != digits.pointer() + digits.length() - 1
	|| n < 0
	|| n > 65535L) {
      mgr.message(PosixHttpStorageMessages::invalidPort,
		  StringMessageArg(id));
      return 0;
    }
    port = (unsigned short)n;
  }
  else
    port = 80;
  String<char> path;
  if (i < id.length()) {
    while (i < id.length() && id[i] != '#') {
      path += char(id[i]);
      i++;
    }
  }
  if (path.length() == 0)
    path += '/';
  return new PosixHttpStorageObject(host, port, path);
}

void PosixHttpStorageManager::mergeId(CString &id, const CString &specId)
{
  size_t i;
  for (i = 0; i < id.length(); i++)
    if (id[i] != '/')
      break;
  size_t slashCount = i;
  if (slashCount > 0) {
    Boolean foundSameSlash = 0;
    size_t sameSlashPos;
    for (size_t j = 0; j < specId.length(); j++) {
      size_t thisSlashCount = 0;
      for (size_t k = j; k < specId.length() && specId[k] == '/'; k++)
	thisSlashCount++;
      if (thisSlashCount == slashCount && !foundSameSlash) {
	foundSameSlash = 1;
	sameSlashPos = j;
      }
      else if (thisSlashCount > slashCount)
	foundSameSlash = 0;
    }
    if (foundSameSlash) {
      CString tem(specId.pointer(), sameSlashPos);
      tem += id;
      tem.moveTo(id);
    }
  }
  else {
    size_t j;
    for (j = specId.length(); j > 0; j--)
      if (specId[j - 1] == '/')
	break;
    if (j > 0) {
      CString tem(specId.pointer(), j);
      tem += id;
      tem.moveTo(id);
    }
  }
  // FIXME remove xxx/../, and /.
}

PosixHttpStorageObject::PosixHttpStorageObject(String<char> &host,
					       unsigned short port,
					       String<char> &path)
: PosixBaseStorageObject(-1), port_(port)
{
  host.moveTo(host_);
  host_ += '\0';
  path.moveTo(path_);
  for (size_t i = 0; i < host_.length() - 1; i++)
    hostStr_ += host_[i];
}

PosixHttpStorageObject::~PosixHttpStorageObject()
{
  if (fd_ > 0)
    (void)xclose(fd_);
}

Boolean PosixHttpStorageObject::open(Messenger &mgr, Boolean mayRewind,
				     size_t &blockSize)
{
  mayRewind_ = mayRewind;

#ifdef HAVE_SOCKET
  if (fd_ < 0) {
    struct sockaddr_in sock;
    sock.sin_family = AF_INET;
    sock.sin_port = htons(port_);
    if (isdigit((unsigned char)host_[0])) {
      unsigned long n = inet_addr(host_.pointer());
      if (n == (unsigned long)-1) {
	ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::invalidHostNumber,
					     StringMessageArg(hostStr_));
	return 0;
      }
      sock.sin_addr.s_addr = n;
    }
    else {
      struct hostent *hp = gethostbyname(host_.pointer());
      if (!hp) {
	// FIXME use h_errno
	ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::cannotResolveHost,
					     StringMessageArg(hostStr_));
	return 0;
      }
      memcpy(&sock.sin_addr, hp->h_addr, hp->h_length);
    }
    fd_ = socket(AF_INET, SOCK_STREAM, 0);
    if (fd_ < 0) {
      ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::cannotCreateSocket,
					   ErrnoMessageArg(errno));
      return 0;
    }
    if (connect(fd_, (struct sockaddr *)&sock, sizeof(sock)) < 0) {
      ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::cannotConnect,
					   StringMessageArg(hostStr_),
					   ErrnoMessageArg(errno));
      (void)xclose(fd_);
      fd_ = -1;
      return 0;
    }
    String<char> request;
    request.append("GET ", 4);
    request += path_;
    request += ' ';
    request.append("HTTP/1.0\r\n\r\n", 12);
    // FIXME check length of write
    if (write(fd_, request.pointer(), request.length()) < 0) {
      ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::writeError,
					   StringMessageArg(hostStr_),
					   ErrnoMessageArg(errno));
      (void)xclose(fd_);
      fd_ = -1;
      return 0;
    }
    if (!readHeader(mgr)) {
      (void)xclose(fd_);
      fd_ = -1;
      return 0;
    }
    rewindPrepare();
  }
  blockSize = getBlockSize();
  return 1;
#else /* not HAVE_SOCKET */
  ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::notSupported);
  return 0;
#endif /* not HAVE_SOCKET */
}

Boolean PosixHttpStorageObject::readHeader(Messenger &mgr)
{
  String<char> buf;
  String<char> leftOver;
  if (!readLine(mgr, buf, leftOver))
    return 0;
  buf += '\0';
  static const char ver[] = "HTTP/1.0";
  int i;
  for (i = 0; ver[i]; i++)
    if (ver[i] != buf[i])
      break;
  if (ver[i]) {
    if (buf.length() > 0)
      unread(buf.pointer(), buf.length() - 1);
    return 1;
  }
  const char *ptr = &buf[i];
  while (isspace((unsigned char)*ptr))
    ptr++;
  int val = 0;
  while (isdigit((unsigned char)*ptr)) {
    val = val*10 + *ptr - '0';
    ptr++;
  }
  while (isspace((unsigned char)*ptr))
    ptr++;
  if (val < 200 || val >= 300) {
    CString reason;
    while (*ptr && *ptr != '\n' && *ptr != '\r') {
      reason += Char(*ptr);
      ptr++;
    }
    CString pathStr;
    for (i = 0; i < path_.length(); i++)
      pathStr += path_[i];
    ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::getFailed,
					 StringMessageArg(hostStr_),
					 StringMessageArg(pathStr),
					 StringMessageArg(reason));
    return 0;
  }
					 
  for (;;) {
    if (!readLine(mgr, buf, leftOver))
      return 0;
    if (buf.length() == 0 || buf[0] == '\r' || buf[0] == '\n')
      break;
  }
  if (leftOver.length())
    unread(leftOver.pointer(), leftOver.length());
  return 1;
}

// True will be returned for an empty line.

Boolean PosixHttpStorageObject::readLine(Messenger &mgr,
					 String<char> &line,
					 String<char> &leftOver)
{
  line.clear();
  Boolean hadCr = 0;
  Boolean gotLine = 0;
  size_t li;
  for (li = 0; li < leftOver.length(); li++) {
    if (leftOver[li] == '\r') {
      if (hadCr) {
	gotLine = 1;
	break;
      }
      line += '\r';
      hadCr = 1;
    }
    else if (leftOver[li] == '\n') {
      line += '\n';
      li++;
      gotLine = 1;
      break;
    }
    else if (hadCr) {
      gotLine = 1;
      break;
    }
    else
      line += leftOver[li];
  }
  if (gotLine) {
    if (li < leftOver.length()) {
      memmove(leftOver.pointer(),
	      leftOver.pointer() + li,
	      leftOver.length() - li);
      leftOver.setLength(leftOver.length() - li);
    }
    else
      leftOver.clear();
    return 1;
  }
  leftOver.clear();
  if (eof_)
    return 1;
  for (;;) {
    char c;
    long n;
    do {
      n = ::read(fd_, &c, 1);
    } while (n < 0 && errno == EINTR);
    if (n == 0) {
      (void)xclose(fd_);
      eof_ = 1;
      return 1;
    }
    if (n < 0) {
      ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::readError,
					   StringMessageArg(hostStr_),
					   ErrnoMessageArg(errno));
      (void)xclose(fd_);
      fd_ = -1;
      return 0;
    }
    switch (c) {
    case '\r':
      if (hadCr) {
	leftOver += c;
	return 1;
      }
      hadCr = 1;
      line += c;
      break;
    case '\n':
      line += c;
      return 1;
    default:
      if (hadCr) {
	leftOver += c;
	return 1;
      }
      line += c;
      break;
    }
  }
  return 0;			// not reached
}

Boolean PosixHttpStorageObject::read(char *buf, size_t bufSize, Messenger &mgr,
				     size_t &nread)
{
  if (readSaved(buf, bufSize, nread))
    return 1;
  if (fd_ < 0 || eof_)
    return 0;
  long n;
  do {
    n = ::read(fd_, buf, bufSize);
  } while (n < 0 && errno == EINTR);
  if (n > 0) {
    nread = size_t(n);
    saveBytes(buf, nread);
    return 1;
  }
  if (n < 0) {
    ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::readError,
					 StringMessageArg(hostStr_),
					 ErrnoMessageArg(errno));
    fd_ = -1;
  }
  else {
    eof_ = 1;
    if (xclose(fd_) < 0)
      ParentLocationMessenger(mgr).message(PosixHttpStorageMessages::closeError,
					   StringMessageArg(hostStr_),
					   ErrnoMessageArg(errno));
    fd_ = -1;
  }
  return 0;
}

Boolean PosixHttpStorageObject::seek(off_t, Messenger &)
{
  CANNOT_HAPPEN();
  return 0;
}

