// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif

#include "config.h"
#include "PosixStorage.H"
#include "PosixBaseStorageObject.H"
#include "StorageManager.H"
#include "DescriptorManager.H"
#include "MessageArg.H"
#include "ErrnoMessageArg.H"
#include "Message.H"
#include "CString.H"
#include "StringOf.H"
#include "CharsetInfo.H"
#include "CodingSystem.H"
#include "macros.H"
#include "PosixStorageMessages.H"

#include <sys/types.h>
#include <stdio.h>

#ifdef INCLUDE_IO_H
#include <io.h>		// for open, fstat, lseek, read prototypes
#endif

#ifdef INCLUDE_UNISTD_H
#include <unistd.h>
#endif

#ifdef INCLUDE_OSFCN_H
#include <osfcn.h>
#endif

#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <stddef.h>

#ifndef S_ISREG
#ifndef S_IFREG
#define S_IFREG _S_IFREG
#endif
#ifndef S_IFFMT
#define S_IFFMT _S_IFFMT
#endif
#define S_ISREG(m) (((m) & S_IFMT) == S_IFREG)
#endif /* not S_ISREG */

#ifndef O_BINARY
#define O_BINARY 0
#endif

PosixBaseStorageObject::PosixBaseStorageObject(int fd)
: fd_(fd), mayRewind_(0), eof_(0), readingSaved_(0), savingBytes_(0)
{
}

void PosixBaseStorageObject::rewindPrepare()
{
  if (!mayRewind_)
    return;
  struct stat sb;
  if (fstat(fd_, &sb) < 0 || !S_ISREG(sb.st_mode)
      || (startOffset_ = lseek(fd_, off_t(0), SEEK_CUR)) < 0) {
    canSeek_ = 0;
    savingBytes_ = 1;
  }
  else
    canSeek_ = 1;
}

Boolean PosixBaseStorageObject::rewind(Messenger &mgr)
{
  ASSERT(mayRewind_);
  if (canSeek_) {
    eof_ = 0;
    return seek(startOffset_, mgr);
  }
  else {
    readingSaved_ = 1;
    nBytesRead_ = 0;
    return 1;
  }
}

void PosixBaseStorageObject::unread(const char *s, size_t n)
{
  savedBytes_.append(s, n);
  if (!readingSaved_) {
    readingSaved_ = 1;
    nBytesRead_ = 0;
  }
}

void PosixBaseStorageObject::willNotRewind()
{
  mayRewind_ = 0;
  savingBytes_ = 0;
  if (!readingSaved_)
    savedBytes_.clear();
  if (eof_ && fd_ >= 0) {
    close();
    fd_ = -1;
  }
}

Boolean PosixBaseStorageObject::readSaved(char *buf, size_t bufSize,
					  size_t &nread)
{
  if (!readingSaved_)
    return 0;
  if (nBytesRead_ >= savedBytes_.length()) {
    if (!mayRewind_)
      savedBytes_.clear();
    readingSaved_ = 0;
    return 0;
  }
  nread = savedBytes_.length() - nBytesRead_;
  if (nread > bufSize)
    nread = bufSize;
  memcpy(buf, savedBytes_.pointer() + nBytesRead_, nread);
  nBytesRead_ += nread;
  return 1;
}

void PosixBaseStorageObject::close()
{
}

int PosixBaseStorageObject::xclose(int fd)
{
  int ret;
  do {
    ret = ::close(fd);
  } while (ret < 0 && errno == EINTR);
  return ret;
}


class PosixStorageObject : public PosixBaseStorageObject, private DescriptorUser {
public:
  PosixStorageObject(const CString &,
		     const OutputCodingSystem *,
		     DescriptorManager *);
  ~PosixStorageObject();
  Boolean open(Messenger &, Boolean mayRewind, size_t &blockSize);
  Boolean read(char *buf, size_t bufSize, Messenger &mgr, size_t &nread);
  Boolean suspend();
  Boolean seek(off_t, Messenger &);
  void close();
private:
  void resume(Messenger &);

  PackedBoolean suspended_;
  off_t suspendPos_;
  const MessageType2 *suspendFailedMessage_;
  int suspendErrno_;
  CString filename_;
  String<char> filenameBytes_;

  void systemError(Messenger &, const MessageType2 &, int);
};

PosixStorageManager::PosixStorageManager(const char *type,
					 const UnivCharsetDesc &filenameCharset,
					 const OutputCodingSystem *filenameCodingSystem,
					 int maxFDs)
: StorageManager(filenameCharset),
  type_(type),
  filenameCodingSystem_(filenameCodingSystem),
  descriptorManager_(maxFDs)
{
  Char newline = idCharset().execToDesc('\n');
  reString_.set(&newline, 1);
}

const char *PosixStorageManager::type() const
{
  return type_;
}


#ifdef POSIX_FILENAMES

#define MERGE_DEFINED

// FIXME should use idCharset.

void PosixStorageManager::mergeId(CString &id, const CString &specId)
{
  if (id.length() == 0 || id[0] == '/')
    return;
  size_t dirLength = 0;
  for (size_t i = 0; i < specId.length(); i++)
    if (specId[i] == '/')
      dirLength = i + 1;
  if (dirLength) {
    CString merged(specId.pointer(), dirLength);
    merged += id;
    merged.moveTo(id);
  }
}

#endif /* POSIX_FILENAMES */

#ifdef MSDOS_FILENAMES

#define MERGE_DEFINED

void PosixStorageManager::mergeId(CString &id, const CString &specId)
{
  if (id.length() == 0 || id[0] == '/' || id[0] == '\\'
      || (id.length() > 0 && id[1] == ':'))
    return;
  size_t dirLength = 0;
  for (size_t i = 0; i < specId.length(); i++)
    if (specId[i] == '/' || specId[i] == '\\'
	|| (i == 1 && specId[i] == ':'))
      dirLength = i + 1;
  if (dirLength) {
    CString merged(specId.pointer(), dirLength);
    merged += id;
    merged.moveTo(id);
  }
}

#endif /* MSDOS_FILENAMES */

#ifndef MERGE_DEFINED

void PosixStorageManager::mergeId(CString &, const CString &)
{
}

#endif /* not MERGE_DEFINED */

String<char> PosixStorageManager::canonId(const CString &id)
{
  String<char> buf = filenameCodingSystem_->convertOut(id);
#ifdef _POSIX_VERSION
  struct stat sb;
  // A regular file is uniquely identifier by st_dev and st_ino.
  if (stat(buf.pointer(), &sb) >= 0 && S_ISREG(sb.st_mode)) {
    buf.setLength(32);
    sprintf(buf.pointer(),
	    "%lu:%lu",
	    (unsigned long)sb.st_dev,
	    (unsigned long)sb.st_ino);
  }
#endif /* _POSIX_VERSION */
  buf.setLength(strlen(buf.pointer()));
  return buf;
}

StorageObject *
PosixStorageManager::makeStorageObject(const CString &str, Messenger &mgr)
{
  if (str.length() == 0) {
    mgr.message(PosixStorageMessages::invalidFilename,
		StringMessageArg(str));
    return 0;
  }
  return new PosixStorageObject(str, filenameCodingSystem_,
				&descriptorManager_);
}

PosixStorageObject::PosixStorageObject(const CString &filename,
				       const OutputCodingSystem *filenameCodingSystem,
				       DescriptorManager *manager)
: DescriptorUser(manager),
  PosixBaseStorageObject(-1),
  suspended_(0),
  filename_(filename),
  filenameBytes_(filenameCodingSystem->convertOut(filename))
{
}

PosixStorageObject::~PosixStorageObject()
{
  if (fd_ >= 0) {
    (void)xclose(fd_);
    releaseD();
  }
}

Boolean PosixStorageObject::seek(off_t off, Messenger &mgr)
{
  if (lseek(fd_, off, SEEK_SET) < 0) {
    fd_ = -1;
    systemError(mgr, PosixStorageMessages::lseekSystemCall, errno);
    return 0;
  }
  else
    return 1;
}

Boolean PosixStorageObject::open(Messenger &mgr, Boolean mayRewind,
				 size_t &blockSize)
{
  mayRewind_ = mayRewind;
  rewindPrepare();
  if (fd_ < 0) {
    acquireD();
    do {
      fd_ = ::open(filenameBytes_.pointer(), O_RDONLY|O_BINARY);
    } while (fd_ < 0 && errno == EINTR);
    if (fd_ < 0) {
      releaseD();
      systemError(mgr, PosixStorageMessages::openSystemCall, errno);
      return 0;
    }
  }
  blockSize = getBlockSize();
  return 1;
}

Boolean PosixStorageObject::read(char *buf, size_t bufSize, Messenger &mgr,
				 size_t &nread)
{
  if (readSaved(buf, bufSize, nread))
    return 1;
  if (suspended_)
    resume(mgr);
  if (fd_ < 0 || eof_)
    return 0;
  long n;
  do {
    n = ::read(fd_, buf, bufSize);
  } while (n < 0 && errno == EINTR);
  if (n > 0) {
    nread = size_t(n);
    saveBytes(buf, nread);
    return 1;
  }
  if (n < 0) {
    int saveErrno = errno;
    releaseD();
    (void)xclose(fd_);
    systemError(mgr, PosixStorageMessages::readSystemCall, saveErrno);
    fd_ = -1;
  }
  else {
    eof_ = 1;
    // n == 0, so end of file
    if (!mayRewind_) {
      releaseD();
      if (xclose(fd_) < 0)
	systemError(mgr, PosixStorageMessages::closeSystemCall, errno);
      fd_ = -1;
    }
    
  }
  return 0;
}

void PosixStorageObject::close()
{
  releaseD();
  (void)xclose(fd_);
}

Boolean PosixStorageObject::suspend()
{
  if (fd_ < 0 || suspended_)
    return 0;
  struct stat sb;
  if (fstat(fd_, &sb) < 0 || !S_ISREG(sb.st_mode))
    return 0;
  suspendFailedMessage_ = 0;
  suspendPos_ = lseek(fd_, 0, SEEK_CUR);
  if (suspendPos_ == (off_t)-1) {
    suspendFailedMessage_ = &PosixStorageMessages::lseekSystemCall;
    suspendErrno_ = errno;
  }
  if (xclose(fd_) < 0 && !suspendFailedMessage_) {
    suspendFailedMessage_ = &PosixStorageMessages::closeSystemCall;
    suspendErrno_ = errno;
  }
  fd_ = -1;
  suspended_ = 1;
  releaseD();
  return 1;
}

void PosixStorageObject::resume(Messenger &mgr)
{
  ASSERT(suspended_);
  if (suspendFailedMessage_) {
    systemError(mgr, *suspendFailedMessage_, suspendErrno_);
    suspended_ = 0;
    return;
  }
  acquireD();
  // suspended_ must be 1 until after acquireD() is called,
  // so that we don't try to suspend this one before it is resumed.
  suspended_ = 0;
  do {
    fd_ = ::open(filenameBytes_.pointer(), O_RDONLY|O_BINARY);
  } while (fd_ < 0 && errno == EINTR);
  if (fd_ < 0) {
    releaseD();
    systemError(mgr, PosixStorageMessages::openSystemCall, errno);
    return;
  }
  if (lseek(fd_, suspendPos_, SEEK_SET) < 0) {
    systemError(mgr, PosixStorageMessages::lseekSystemCall, errno);
    (void)xclose(fd_);
    fd_ = -1;
    releaseD();
  }
}

#ifdef STAT_BLKSIZE

size_t PosixBaseStorageObject::getBlockSize() const
{
  struct stat sb;
  long sz;
  if (fstat(fd_, &sb) < 0)
    return defaultBlockSize;
  if (!S_ISREG(sb.st_mode))
    return defaultBlockSize;
  if (sb.st_blksize > size_t(-1))
    sz = size_t(-1);
  else
    sz = sb.st_blksize;
  return sz;
}

#else /* not STAT_BLKSIZE */

size_t PosixBaseStorageObject::getBlockSize() const
{
  return defaultBlockSize;
}

#endif /* not STAT_BLKSIZE */

void PosixStorageObject::systemError(Messenger &mgr,
				     const MessageType2 &msg,
				     int err)
{
  ParentLocationMessenger(mgr).message(msg,
				       StringMessageArg(filename_),
				       ErrnoMessageArg(err));
}

class PosixFdStorageObject : public PosixBaseStorageObject {
public:
  PosixFdStorageObject(int);
  Boolean open(Messenger &, Boolean mayRewind, size_t &blockSize);
  Boolean read(char *buf, size_t bufSize, Messenger &mgr, size_t &nread);
  Boolean seek(off_t, Messenger &);
  enum {
    noError,
    readError,
    invalidNumberError,
    lseekError
  };
private:
  int origFd_;
};

PosixFdStorageManager::PosixFdStorageManager(const char *type,
					     const UnivCharsetDesc &idCharset)
: StorageManager(idCharset), type_(type)
{
}

StorageObject *PosixFdStorageManager::makeStorageObject(const CString &id,
							Messenger &mgr)
{
  int n = 0;
  for (int i = 0; i < id.length(); i++) {
    UnivChar ch;
    if (!idCharset().descToUniv(id[i], ch))
      break;
    if (ch < UnivCharsetDesc::zero || ch > UnivCharsetDesc::zero + 9)
      break;
    int digit = ch - UnivCharsetDesc::zero;
    // Allow the division to be done at compile-time.
    if (n > INT_MAX/10)
      break;
    n *= 10;
    if (n > INT_MAX - digit)
      break;
    n += digit;
  }
  if (i < id.length() || i == 0) {
    mgr.message(PosixStorageMessages::invalidNumber,
		StringMessageArg(id));
    return 0;
  }
  return new PosixFdStorageObject(n);
}

PosixFdStorageObject::PosixFdStorageObject(int fd)
: PosixBaseStorageObject(fd), origFd_(fd)
{
}

const char *PosixFdStorageManager::type() const
{
  return type_;
}

Boolean PosixFdStorageObject::open(Messenger &, Boolean mayRewind,
				   size_t &blockSize)
{
  mayRewind_ = mayRewind;
  rewindPrepare();
  // Could check access mode with fcntl(fd_, F_GETFL).
  blockSize = getBlockSize();
  return 1;
}

Boolean PosixFdStorageObject::read(char *buf, size_t bufSize, Messenger &mgr,
				   size_t &nread)
{
  if (readSaved(buf, bufSize, nread))
    return 1;
  if (fd_ < 0 || eof_)
    return 0;
  long n;
  do {
    n = ::read(fd_, buf, bufSize);
  } while (n < 0 && errno == EINTR);
  if (n > 0) {
    nread = size_t(n);
    saveBytes(buf, nread);
    return 1;
  }
  if (n < 0) {
    ParentLocationMessenger(mgr).message(PosixStorageMessages::fdRead,
					 NumberMessageArg(fd_),
					 ErrnoMessageArg(errno));
    fd_ = -1;
  }
  else
    eof_ = 1;
  return 0;
}

Boolean PosixFdStorageObject::seek(off_t off, Messenger &mgr)
{
  if (lseek(fd_, off, SEEK_SET) < 0) {
    ParentLocationMessenger(mgr).message(PosixStorageMessages::fdLseek,
					 NumberMessageArg(fd_),
					 ErrnoMessageArg(errno));
    return 0;
  }
  else
    return 1;
}
