// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif

#include "config.h"
#include "StdioStorage.H"
#include "Message.H"
#include "types.H"
#include "ErrnoMessageArg.H"
#include "StringOf.H"
#include "CString.H"
#include "CodingSystem.H"

#include "StdioStorageMessages.H"

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <errno.h>

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

class StdioStorageObject : public StorageObject {
public:
  StdioStorageObject(const CString &filename,
		     const OutputCodingSystem *filenameCodingSystem);
  ~StdioStorageObject();
  Boolean open(Messenger &, Boolean mayRewind, size_t &blockSize);
  Boolean read(char *buf, size_t bufSize, Messenger &mgr, size_t &nread);
  Boolean rewind(Messenger &mgr);
private:
  enum ErrorIndex {
    invalidErrorIndex,
    fopenFailed,
    readError,
    seekError
  };
  void error(Messenger &mgr, const MessageType2 &, int err);

  FILE *fp_;
  CString filename_;
  String<char> filenameBytes_;
};

StdioStorageManager::StdioStorageManager(const char *type,
					 const UnivCharsetDesc &filenameCharset,
					 const OutputCodingSystem *filenameCodingSystem)
: StorageManager(filenameCharset),
  type_(type),
  filenameCodingSystem_(filenameCodingSystem)
{
}

StorageObject *StdioStorageManager::makeStorageObject(const CString &str,
						      Messenger &)
{
  return new StdioStorageObject(str, filenameCodingSystem_);
}

const char *StdioStorageManager::type() const
{
  return type_;
}

StdioStorageObject::StdioStorageObject(const CString &filename,
				       const OutputCodingSystem *filenameCodingSystem)
: fp_(0),
  filename_(filename),
  filenameBytes_(filenameCodingSystem->convertOut(filename))
{
}

StdioStorageObject::~StdioStorageObject()
{
  if (fp_) {
    fclose(fp_);
    fp_ = 0;
  }
}

Boolean StdioStorageObject::rewind(Messenger &mgr)
{
  if (fp_) {
    errno = 0;
    if (fseek(fp_, 0L, SEEK_SET) < 0) {
      error(mgr, StdioStorageMessages::seekFailed, errno);
      return 0;
    }
    return 1;
  }
  return 1;
}

Boolean StdioStorageObject::open(Messenger &mgr, Boolean,
				 size_t &blockSize)
{
  if (!fp_) {
    errno = 0;
    fp_ = fopen(filenameBytes_.pointer(), "r");
    if (!fp_) {
      error(mgr, StdioStorageMessages::openFailed, errno);
      return 0;
    }
  }
  blockSize = BUFSIZ;
  return 1;
}

Boolean StdioStorageObject::read(char *buf, size_t bufSize, Messenger &mgr,
				 size_t &nread)
{
  if (!fp_)
    return 0;
  errno = 0;
  size_t n = 0;
  FILE *fp = fp_;
  while (n < bufSize) {
    int c = getc(fp);
    if (c == EOF) {
      if (ferror(fp)) {
	error(mgr, StdioStorageMessages::readFailed, errno);
	(void)fclose(fp);
	return 0;
      }
      fclose(fp);
      fp_ = 0;
      break;
    }
    buf[n++] = c;
  }
  nread = n;
  return n > 0;
}

void StdioStorageObject::error(Messenger &mgr,
			       const MessageType2 &msg,
			       int err)
{
  ParentLocationMessenger(mgr).message(msg,
				       StringMessageArg(filename_),
				       ErrnoMessageArg(err));
}
