// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifndef Event_INCLUDED
#define Event_INCLUDED 1
#ifdef __GNUG__
#pragma interface
#endif

#include "Link.H"
#include "Allocator.H"
#include "Location.H"
#include "Vector.H"
#include "Owner.H"
#include "Boolean.H"
#include "types.H"
#include "ResourcePointer.H"
#include "CString.H"
#include "Notation.H"
#include "Sd.H"
#include "Syntax.H"
#include "Dtd.H"
#include "ElementType.H"
#include "Text.H"
#include "Lpd.H"
#include "Message.H"
#include "Decl.H"

class EventHandler;

class Event : public Link {
public:
  enum Type {
    message,
    characterData,
    startElement,
    endElement,
    pi,
    sdataEntity,
    externalDataEntity,
    subdocEntity,
    appinfo,
    startDtd,
    endDtd,
    endProlog,
    sgmlDecl,
    simpleLink,
    complexLink,
    uselink,
    usemap,
    commentDecl,
    sSep,
    ignoredRs,
    ignoredRe,
    reOrigin,
    ignoredChars,
    markedSectionStart,
    markedSectionEnd,
    entityStart,
    entityEnd,
    notationDecl,
    externalEntityDecl
    };
  Event(Type);
  virtual ~Event();
  virtual void handle(EventHandler &) = 0;
  virtual void copyData();
  void *operator new(size_t sz, Allocator &alloc) { return alloc.alloc(sz); }
  void *operator new(size_t sz) { return Allocator::allocSimple(sz); }
  void operator delete(void *p) { Allocator::free(p); }
  Type type() const;
private:
  Event(const Event &);		// undefined
  void operator=(const Event &); // undefined
  Type type_;
};

class LocatedEvent : public Event {
public:
  LocatedEvent(Type type, const Location &);
  const Location &location() const;
private:
  LocatedEvent(const LocatedEvent &); // undefined
  void operator=(const LocatedEvent &);	// undefined
  Location location_;
};

class MessageEvent : public Event {
public:
  MessageEvent(Message &);
  const Message &message() const;
  void handle(EventHandler &);
private:
  MessageEvent(const MessageEvent &); // undefined
  void operator=(const MessageEvent &);	// undefined
  Message message_;
};

class ElementEvent : public LocatedEvent {
public:
  enum Minimization {
    complete,
    unclosed,
    empty,
    net,
    omittedOmittag,
    omittedEmpty,
    omittedConref
    };
  ElementEvent(Type type,
	       const ElementType *,
	       const ConstResourcePointer<Dtd> &,
	       const Location &,
	       Index nextIndex,
	       Minimization, CString &, size_t);
  ElementEvent(Type type,
	       const ElementType *,
	       const ConstResourcePointer<Dtd> &,
	       const Location &,
	       Index nextIndex,
	       Minimization);
  Index nextIndex() const;	// index in entity of next character after tag
  void setIncluded();
  Boolean included() const;
  Minimization minimization() const;
  const CString &name() const;
  const CString origName() const;
  const CString finalS() const;
  const ElementType *elementType() const;
protected:
  CString origChars_;
private:
  ElementEvent(const ElementEvent &); // undefined
  void operator=(const ElementEvent &);	// undefined
  const ElementType *elementType_;
  ConstResourcePointer<Dtd> dtd_;
  Minimization minimization_;
  PackedBoolean included_;
  PackedBoolean copied_;	// has name_ been copied?
  Index nextIndex_;
  size_t origNameLength_;
};

class AttributeList;

class StartElementEvent : public ElementEvent {
public:
  StartElementEvent(const ElementType *,
		    const ConstResourcePointer<Dtd> &,
		    const Location &,
		    Index nextIndex,
		    Minimization,
		    AttributeList *,
		    CString &,
		    size_t);
  StartElementEvent(const ElementType *,
		    const ConstResourcePointer<Dtd> &,
		    const Location &,
		    Index nextIndex,
		    Minimization,
		    AttributeList *);
  ~StartElementEvent();
  void handle(EventHandler &);
  Boolean netEnabling() const;
  const AttributeList &attributes() const;
  void copyData();
private:
  StartElementEvent(const StartElementEvent &);	// undefined
  void operator=(const StartElementEvent &);	// undefined
  AttributeList *attributes_;
  PackedBoolean copied_;
};

class EndElementEvent : public ElementEvent {
public:
  EndElementEvent(const ElementType *,
		  const ConstResourcePointer<Dtd> &,
		  const Location &, Index nextIndex,
		  Minimization, CString &, size_t);
  EndElementEvent(const ElementType *,
		  const ConstResourcePointer<Dtd> &,
		  const Location &, Index nextIndex,
		  Minimization);

  void handle(EventHandler &);
  const CString &origChars() const;
private:
  EndElementEvent(const EndElementEvent &); // undefined
  void operator=(const EndElementEvent &);  // undefined
};

class DataEvent : public LocatedEvent {
public:
  DataEvent(Type, const Char *, size_t, const Location &);
  void handle(EventHandler &);
  const Char *data() const;
  size_t dataLength() const;
  virtual Boolean isRe(unsigned long &serial) const;
  virtual const Entity *entity() const;
protected:
  const Char *p_;
  size_t length_;
private:
  DataEvent(const DataEvent &);	// undefined
  void operator=(const DataEvent &); // undefined
};

class ImmediateDataEvent : public DataEvent {
public:
  ImmediateDataEvent(Type type, const Char *, size_t, const Location &,
		     Boolean copy);
  ~ImmediateDataEvent();
  void copyData();
private:
  ImmediateDataEvent(const ImmediateDataEvent &); // undefined
  void operator=(const ImmediateDataEvent &);	  // undefined
  Char *alloc_;
};

class InternalDataEntity;

class DataEntityEvent : public DataEvent {
public:
  DataEntityEvent(Type type, const InternalDataEntity *,
		  const ConstResourcePointer<Origin> &);
  const Entity *entity() const;
private:
  DataEntityEvent(const DataEntityEvent &); // undefined
  void operator=(const DataEntityEvent &);  // undefined
};

class InternalCdataEntity;

class CdataEntityEvent : public DataEntityEvent {
public:
  CdataEntityEvent(const InternalCdataEntity *,
		   const ConstResourcePointer<Origin> &);
private:
  CdataEntityEvent(const CdataEntityEvent &); // undefined
  void operator=(const CdataEntityEvent &);   // undefined
};

class InternalSdataEntity;

class SdataEntityEvent : public DataEntityEvent {
public:
  SdataEntityEvent(const InternalSdataEntity *,
		   const ConstResourcePointer<Origin> &);
  void handle(EventHandler &);
private:
  SdataEntityEvent(const SdataEntityEvent &); // undefined
  void operator=(const SdataEntityEvent &);   // undefined
};

class PiEntity;

class PiEvent : public LocatedEvent {
public:
  PiEvent(const Char *, size_t, const Location &);
  const Char *data() const;
  size_t dataLength() const;
  virtual const Entity *entity() const;
  void handle(EventHandler &);
private:
  PiEvent(const PiEvent &);	// undefined
  void operator=(const PiEvent &); // undefined
  const Char *data_;
  size_t dataLength_;
};

class ImmediatePiEvent : public PiEvent {
public:
  ImmediatePiEvent(CString &, const Location &);
private:
  ImmediatePiEvent(const ImmediatePiEvent &); // undefined
  void operator=(const ImmediatePiEvent &);   // undefined
  CString string_;
};

class PiEntityEvent : public PiEvent {
public:
  PiEntityEvent(const PiEntity *entity,
		const ConstResourcePointer<Origin> &origin);
  const Entity *entity() const;
private:
  PiEntityEvent(const PiEntityEvent &);	// undefined
  void operator=(const PiEntityEvent &); // undefined
};

class ExternalNonTextEntity;
class ExternalDataEntity;
class SubdocEntity;

class ExternalEntityEvent : public Event {
public:
  ExternalEntityEvent(Type type,
		      const ConstResourcePointer<EntityOrigin> &);
  const ConstResourcePointer<EntityOrigin> &entityOrigin() const;
  const Location &location() const;
private:
  ExternalEntityEvent(const ExternalEntityEvent &); // undefined
  void operator=(const ExternalEntityEvent &);	    // undefined
  ConstResourcePointer<EntityOrigin> origin_;
};

class ExternalDataEntityEvent : public ExternalEntityEvent {
public:
  ExternalDataEntityEvent(const ExternalDataEntity *,
			  const ConstResourcePointer<EntityOrigin> &);
  void handle(EventHandler &);
  const ExternalDataEntity *entity() const;
private:
  ExternalDataEntityEvent(const ExternalDataEntityEvent &); // undefined
  void operator=(const ExternalDataEntityEvent &);	    // undefined
  const ExternalDataEntity *dataEntity_;
};

class SubdocEntityEvent : public ExternalEntityEvent {
public:
  SubdocEntityEvent(const SubdocEntity *,
		    const ConstResourcePointer<EntityOrigin> &);
  void handle(EventHandler &);
  const SubdocEntity *entity() const;
private:
  SubdocEntityEvent(const SubdocEntityEvent &);	// undefined
  void operator=(const SubdocEntityEvent &);	// undefined
  const SubdocEntity *subdocEntity_;
};

class AppinfoEvent : public LocatedEvent {
public:
  AppinfoEvent(const Location &);
  AppinfoEvent(const Text &, const Location &);
  void handle(EventHandler &);
  Boolean literal(const CString *&) const;
private:
  AppinfoEvent(const AppinfoEvent &); // undefined
  void operator=(const AppinfoEvent &);	// undefined
  Boolean appinfoNone_;
  Text appinfo_;
};

class SimpleLinkEvent : public LocatedEvent {
public:
  SimpleLinkEvent(const CString &,
		  AttributeList &,
		  const Location &);
  void handle(EventHandler &);
  const AttributeList &attributes() const;
  const CString &name() const;
private:
  SimpleLinkEvent(const SimpleLinkEvent &); // undefined
  void operator=(const SimpleLinkEvent &);
  CString name_;
  AttributeList attributes_;
};

class ComplexLinkEvent : public Event {
public:
  ComplexLinkEvent(const ConstResourcePointer<ComplexLpd> &,
		   const VectorBase<ConstResourcePointer<Lpd> > &activeLpds);
  void handle(EventHandler &);
  const CString &name() const;
  const ConstResourcePointer<ComplexLpd> &lpd() const;
  const VectorBase<ConstResourcePointer<Lpd> > &activeLpds() const;
private:
  ComplexLinkEvent(const ComplexLinkEvent &); // undefined
  void operator=(const ComplexLinkEvent &);
  ConstResourcePointer<ComplexLpd> lpd_;
  Vector<ConstResourcePointer<Lpd> > activeLpds_;
};

class UselinkEvent : public LocatedEvent {
public:
  UselinkEvent(Decl &,
	       const ConstResourcePointer<Lpd> &,
	       const LinkSet *,
	       Boolean restore,
	       const Location &);
  void handle(EventHandler &);
  const ConstResourcePointer<Lpd> &lpd() const;
  const LinkSet *linkSet() const;
  Boolean restore() const;
  const Decl &decl() const;
private:
  UselinkEvent(const UselinkEvent &); // undefined
  void operator=(const UselinkEvent &);	// undefined
  ConstResourcePointer<Lpd> lpd_;
  const LinkSet *linkSet_;
  Boolean restore_;
  Decl decl_;
};

class UsemapEvent : public LocatedEvent {
public:
  UsemapEvent(Decl &,
	      const ShortReferenceMap *,
	      const ConstResourcePointer<Dtd> &,
	      const Location &);
  void handle(EventHandler &);
  const ShortReferenceMap *map() const;
  const Decl &decl() const;
private:
  UsemapEvent(const UsemapEvent &); // undefined
  void operator=(const UsemapEvent &); // undefined
  ConstResourcePointer<Dtd> dtd_;
  const ShortReferenceMap *map_;
  Decl decl_;
};

class StartDtdEvent : public LocatedEvent {
public:
  StartDtdEvent(const CString &,
		const ConstResourcePointer<Entity> &entity,
		const Location &);
  void handle(EventHandler &);
  const CString &name() const;
  const ConstResourcePointer<Entity> &entity() const;
private:
  StartDtdEvent(const StartDtdEvent &);	// undefined
  void operator=(const StartDtdEvent &); // undefined
  CString name_;
  ConstResourcePointer<Entity> entity_;
};

class EndDtdEvent : public LocatedEvent {
public:
  EndDtdEvent(const ConstResourcePointer<Dtd> &, const Location &);
  void handle(EventHandler &);
  const Dtd &dtd() const;
private:
  EndDtdEvent(const EndDtdEvent &); // undefined
  void operator=(const EndDtdEvent &); // undefined
  ConstResourcePointer<Dtd> dtd_;
};

class EndPrologEvent : public LocatedEvent {
public:
  EndPrologEvent(const Location &);
  void handle(EventHandler &);
private:
  EndPrologEvent(const EndPrologEvent &); // undefined
  void operator=(const EndPrologEvent &); // undefined
};

class SgmlDeclEvent : public LocatedEvent {
public:
  // for an implied SGML declaration
  SgmlDeclEvent(const ConstResourcePointer<Sd> &,
		const ConstResourcePointer<Syntax> &syntax);
  // for an explicit SGML declaration
  SgmlDeclEvent(const ConstResourcePointer<Sd> &,
		const ConstResourcePointer<Syntax> &syntax,
		const ConstResourcePointer<Syntax> &instanceSyntax,
		CString &preS,
		const Location &startLocation,
		Index nextIndex);
  void handle(EventHandler &);
  const Sd &sd() const;
  const ConstResourcePointer<Sd> &sdPointer() const;
  const Syntax &prologSyntax() const;
  const ConstResourcePointer<Syntax> &prologSyntaxPointer() const;
  const Syntax &instanceSyntax() const;
  const ConstResourcePointer<Syntax> &instanceSyntaxPointer() const;
  const CString &preS() const;
private:
  SgmlDeclEvent(const SgmlDeclEvent &);	// undefined
  void operator=(const SgmlDeclEvent &); // undefined
  ConstResourcePointer<Sd> sd_;
  ConstResourcePointer<Syntax> prologSyntax_;
  ConstResourcePointer<Syntax> instanceSyntax_;
  CString preS_;
  Index nextIndex_;
};

class CommentDeclEvent : public LocatedEvent {
public:
  CommentDeclEvent(Decl &, const Location &loc);
  void handle(EventHandler &);
  const Decl &decl() const;
private:
  CommentDeclEvent(const CommentDeclEvent &); // undefined
  void operator=(const CommentDeclEvent &);   // undefined
  Decl decl_;
};

class SSepEvent : public ImmediateDataEvent {
public:
  SSepEvent(const Char *, size_t, const Location &, Boolean copy);
  void handle(EventHandler &);
private:
  SSepEvent(const SSepEvent &);	// undefined
  void operator=(const SSepEvent &); // undefined
};

class IgnoredRsEvent : public LocatedEvent {
public:
  IgnoredRsEvent(Char c, const Location &);
  void handle(EventHandler &);
  Char rs() const;
private:
  IgnoredRsEvent(const IgnoredRsEvent &); // undefined
  void operator=(const IgnoredRsEvent &); // undefined
  Char c_;
};

class IgnoredReEvent : public LocatedEvent {
public:
  IgnoredReEvent(Char c, const Location &, unsigned long serial);
  void handle(EventHandler &);
  Char re() const;
  unsigned long serial() const;
private:
  IgnoredReEvent(const IgnoredReEvent &); // undefined
  void operator=(const IgnoredReEvent &); // undefined
  unsigned long serial_;
  Char c_;
};

class ReEvent : public ImmediateDataEvent {
public:
  ReEvent(const Char *, const Location &, unsigned long serial);
  Boolean isRe(unsigned long &serial) const;
private:
  ReEvent(const ReEvent &);	// undefined
  void operator=(const ReEvent &); // undefined
  unsigned long serial_;
};

class ReOriginEvent : public LocatedEvent {
public:
  ReOriginEvent(Char c, const Location &, unsigned long serial);
  void handle(EventHandler &);
  Char re() const;
  unsigned long serial() const;
private:
  ReOriginEvent(const ReOriginEvent &);	// undefined
  void operator=(const ReOriginEvent &); // undefined
  unsigned long serial_;
  Char c_;
};

class IgnoredCharsEvent : public ImmediateDataEvent {
public:
  IgnoredCharsEvent(const Char *, size_t, const Location &, Boolean copy);
  void handle(EventHandler &);
private:
  IgnoredCharsEvent(const IgnoredCharsEvent &);	// undefined
  void operator=(const IgnoredCharsEvent &);	// undefined
};

class MarkedSectionEvent : public LocatedEvent {
public:
  enum Status { include, rcdata, cdata, ignore }; // in priority order
  MarkedSectionEvent(Type, Status, const Location &);
  Status status() const;
private:
  MarkedSectionEvent(const MarkedSectionEvent &); // undefined
  void operator=(const MarkedSectionEvent &);	  // undefined
  Status status_;
};

class MarkedSectionStartEvent : public MarkedSectionEvent {
public:
  MarkedSectionStartEvent(Decl &, Status, const Location &);
  void handle(EventHandler &);
  const Decl &decl() const;
private:
  MarkedSectionStartEvent(const MarkedSectionStartEvent &); // undefined
  void operator=(const MarkedSectionStartEvent &);	    // undefined
  Decl decl_;
};

class MarkedSectionEndEvent : public MarkedSectionEvent {
public:
  MarkedSectionEndEvent(Status, const Location &);
  void handle(EventHandler &);
private:
  MarkedSectionEndEvent(const MarkedSectionEndEvent &); // undefined
  void operator=(const MarkedSectionEndEvent &);	    // undefined
};

class EntityStartEvent : public Event {
public:
  EntityStartEvent(const ConstResourcePointer<EntityOrigin> &origin);
  void handle(EventHandler &);
  const Entity *entity() const;
  const ConstResourcePointer<EntityOrigin> &entityOrigin() const;
private:
  EntityStartEvent(const EntityStartEvent &); // undefined
  void operator=(const EntityStartEvent &); // undefined

  ConstResourcePointer<EntityOrigin> origin_;
};

class EntityEndEvent : public LocatedEvent {
public:
  EntityEndEvent(const Location &);
  void handle(EventHandler &);
private:
  EntityEndEvent(const EntityEndEvent &); // undefined
  void operator=(const EntityEndEvent &); // undefined
};

class ExternalEntity;

class ExternalEntityDeclEvent : public Event {
public:
  ExternalEntityDeclEvent(const ConstResourcePointer<Entity> &,
			  Boolean duplicate);
  void handle(EventHandler &);
  const ExternalEntity &externalEntity() const;
  const ConstResourcePointer<Entity> &entityPointer() const;
  Boolean duplicate() const;
  // The name of the entity will be empty if this is the default entity.
private:
  Boolean duplicate_;
  // This will actually point to an external entity.
  ConstResourcePointer<Entity> entity_;
};

class NotationDeclEvent : public Event {
public:
  NotationDeclEvent(const ConstResourcePointer<Notation> &);
  void handle(EventHandler &);
  const Notation &notation() const;
  const ConstResourcePointer<Notation> &notationPointer() const;
private:
  NotationDeclEvent(const NotationDeclEvent &);	// undefined
  void operator=(const NotationDeclEvent &);	// undefined
  ConstResourcePointer<Notation> notation_;
};

class EventHandler {
public:
  virtual ~EventHandler();
  virtual void message(MessageEvent *) = 0;
  virtual void data(DataEvent *);
  virtual void startElement(StartElementEvent *);
  virtual void endElement(EndElementEvent *);
  virtual void pi(PiEvent *);
  virtual void sdataEntity(SdataEntityEvent *);
  virtual void externalDataEntity(ExternalDataEntityEvent *);
  virtual void subdocEntity(SubdocEntityEvent *);
  virtual void appinfo(AppinfoEvent *);
  virtual void simpleLink(SimpleLinkEvent *);
  virtual void complexLink(ComplexLinkEvent *);
  virtual void uselink(UselinkEvent *);
  virtual void usemap(UsemapEvent *);
  virtual void startDtd(StartDtdEvent *);
  virtual void endDtd(EndDtdEvent *);
  virtual void endProlog(EndPrologEvent *);
  virtual void sgmlDecl(SgmlDeclEvent *);
  virtual void commentDecl(CommentDeclEvent *);
  virtual void sSep(SSepEvent *);
  virtual void ignoredRs(IgnoredRsEvent *);
  virtual void ignoredRe(IgnoredReEvent *);
  virtual void reOrigin(ReOriginEvent *);
  virtual void ignoredChars(IgnoredCharsEvent *);
  virtual void markedSectionStart(MarkedSectionStartEvent *);
  virtual void markedSectionEnd(MarkedSectionEndEvent *);
  virtual void entityStart(EntityStartEvent *);
  virtual void entityEnd(EntityEndEvent *);
  virtual void notationDecl(NotationDeclEvent *);
  virtual void externalEntityDecl(ExternalEntityDeclEvent *);
};

inline
Event::Event(Type type)
: type_(type)
{
}

inline
Event::Type Event::type() const
{
  return type_;
}

inline
const Location &LocatedEvent::location() const
{
  return location_;
}

inline
const Message &MessageEvent::message() const
{
  return message_;
}

inline
const AttributeList &StartElementEvent::attributes() const
{
  return *attributes_;
}

inline
const ElementType *ElementEvent::elementType() const
{
  return elementType_;
}

inline
const CString &ElementEvent::name() const
{
  return elementType_->name();
}

inline
const CString ElementEvent::origName() const
{
  return CString(origChars_.pointer(), origNameLength_);
}

inline
const CString ElementEvent::finalS() const
{
  return CString(origChars_.pointer() + origNameLength_,
		 origChars_.length() - origNameLength_);
}

inline
void ElementEvent::setIncluded()
{
  included_ = 1;
}

inline
Index ElementEvent::nextIndex() const
{
  return nextIndex_;
}

inline
Boolean ElementEvent::included() const
{
  return included_;
}

inline
ElementEvent::Minimization ElementEvent::minimization() const
{
  return minimization_;
}

inline
Boolean StartElementEvent::netEnabling() const
{
  return minimization() == net;
}

inline
const CString &EndElementEvent::origChars() const
{
  return origChars_;
}

inline
const Char *DataEvent::data() const
{
  return p_;
}

inline
size_t DataEvent::dataLength() const
{
  return length_;
}

inline
const Char *PiEvent::data() const
{
  return data_;
}

inline
size_t PiEvent::dataLength() const
{
  return dataLength_;
}

inline
const ConstResourcePointer<EntityOrigin> &
ExternalEntityEvent::entityOrigin() const
{
  return origin_;
}

inline
const Location &ExternalEntityEvent::location() const
{
  return origin_->parent();
}

inline
const ExternalDataEntity *ExternalDataEntityEvent::entity() const
{
  return dataEntity_;
}

inline
const SubdocEntity *SubdocEntityEvent::entity() const
{
  return subdocEntity_;
}

inline
const AttributeList &SimpleLinkEvent::attributes() const
{
  return attributes_;
}

inline
const CString &SimpleLinkEvent::name() const
{
  return name_;
}

inline
const CString &ComplexLinkEvent::name() const
{
  return lpd_->name();
}

inline
const ConstResourcePointer<ComplexLpd> &ComplexLinkEvent::lpd() const
{
  return lpd_;
}

inline
const VectorBase<ConstResourcePointer<Lpd> > &
ComplexLinkEvent::activeLpds() const
{
  return activeLpds_;
}

inline
const Decl &UselinkEvent::decl() const
{
  return decl_;
}

inline
const ConstResourcePointer<Lpd> &UselinkEvent::lpd() const
{
  return lpd_;
}

inline
const LinkSet *UselinkEvent::linkSet() const
{
  return linkSet_;
}

inline
Boolean UselinkEvent::restore() const
{
  return restore_;
}

inline
const Decl &UsemapEvent::decl() const
{
  return decl_;
}

inline
const ShortReferenceMap *UsemapEvent::map() const
{
  return map_;
}

inline
const CString &StartDtdEvent::name() const
{
  return name_;
}

inline
const ConstResourcePointer<Entity> &StartDtdEvent::entity() const
{
  return entity_;
}

inline
const Dtd &EndDtdEvent::dtd() const
{
  return *dtd_;
}

inline
const Sd &SgmlDeclEvent::sd() const
{
  return *sd_;
}

inline
const ConstResourcePointer<Sd> &SgmlDeclEvent::sdPointer() const
{
  return sd_;
}

inline
const Syntax &SgmlDeclEvent::prologSyntax() const
{
  return *prologSyntax_;
}

inline
const ConstResourcePointer<Syntax> &SgmlDeclEvent::prologSyntaxPointer() const
{
  return prologSyntax_;
}

inline
const Syntax &SgmlDeclEvent::instanceSyntax() const
{
  return *instanceSyntax_;
}

inline
const ConstResourcePointer<Syntax> &SgmlDeclEvent::instanceSyntaxPointer() const
{
  return instanceSyntax_;
}

inline
const CString &SgmlDeclEvent::preS() const
{
  return preS_;
}

inline
const Decl &CommentDeclEvent::decl() const
{
  return decl_;
}

inline
Char IgnoredRsEvent::rs() const
{
  return c_;
}

inline
Char IgnoredReEvent::re() const
{
  return c_;
}

inline
unsigned long IgnoredReEvent::serial() const
{
  return serial_;
}

inline
Char ReOriginEvent::re() const
{
  return c_;
}

inline
unsigned long ReOriginEvent::serial() const
{
  return serial_;
}

inline
MarkedSectionEvent::Status MarkedSectionEvent::status() const
{
  return status_;
}

inline
const Decl &MarkedSectionStartEvent::decl() const
{
  return decl_;
}

inline
const Entity *EntityStartEvent::entity() const
{
  return origin_->entity().pointer();
}

inline
const ConstResourcePointer<EntityOrigin> &
EntityStartEvent::entityOrigin() const
{
  return origin_;
}

inline
const ConstResourcePointer<Entity> &ExternalEntityDeclEvent::entityPointer() const
{
  return entity_;
}

inline
const ExternalEntity &ExternalEntityDeclEvent::externalEntity() const
{
  return *(const ExternalEntity *)entity_.pointer();
}

inline
Boolean ExternalEntityDeclEvent::duplicate() const
{
  return duplicate_;
}

inline
const Notation &NotationDeclEvent::notation() const
{
  return *notation_;
}

inline
const ConstResourcePointer<Notation> &NotationDeclEvent::notationPointer() const
{
  return notation_;
}

#endif /* not Event_INCLUDED */
