// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifndef Location_INCLUDED
#define Location_INCLUDED 1
#ifdef __GNUG__
#pragma interface
#endif

#include "types.H"
#include "Boolean.H"
#include "ResourcePointer.H"
#include "Resource.H"
#include "Boolean.H"
#include "Location.H"
#include "GrowableVector.H"
#include "Owner.H"
#include "Allocator.H"
#include "CString.H"
#include "rtti.H"

class ExternalInfo;
class EntityOrigin;
class Entity;
class Location;

class Origin : public Resource {
public:
  enum RefEndType {
    endOmitted,
    endRE,
    endRefc
    };
  virtual ~Origin();
  virtual const EntityOrigin *asEntityOrigin() const;
  virtual const Location &parent() const = 0;
  virtual Index refLength() const;
  virtual Boolean origChar(Char &) const;
  virtual Boolean inBracketedTextOpenDelim() const;
  virtual Boolean inBracketedTextCloseDelim() const;
  virtual Boolean isNumericCharRef(Origin::RefEndType &refEndType) const;
};

class Location {
public:
  Location();
  Location(Origin *, Index);
  Location(ConstResourcePointer<Origin>, Index);
  void operator+=(Index i) { index_ += i; }
  Index index() const { return index_; }
  const ConstResourcePointer<Origin> &origin() const { return origin_; }
  void clear() { origin_.clear(); }
private:
  ConstResourcePointer<Origin> origin_;
  Index index_;
};

class ExternalInfo {
  RTTI_CLASS
public:
  virtual ~ExternalInfo();
};

class NamedCharRefBase {
public:
  NamedCharRefBase();
  NamedCharRefBase(Index, Origin::RefEndType);
  Index refStartIndex() const;
  Origin::RefEndType refEndType() const;
private:
  Index refStartIndex_;
  Origin::RefEndType refEndType_;
};

class NamedCharRef : public NamedCharRefBase {
public:
  NamedCharRef();
  NamedCharRef(Index, Origin::RefEndType, const CString &origName);
  NamedCharRef(const NamedCharRefBase &, const Char *, size_t);
  const CString &origName() const;
private:
  CString origName_;
};

struct EntityOriginNamedCharRef {
  EntityOriginNamedCharRef();
  EntityOriginNamedCharRef(Index i, const NamedCharRefBase &r, size_t);
  Index replacementIndex;
  size_t origNameOffset;
  NamedCharRefBase ref;
};

class EntityOrigin : public Origin {
public:
  enum IndexType {
    normalCharIndex,
    replacementCharIndex
  };
  void *operator new(size_t sz, Allocator &alloc) {
    return alloc.alloc(sz);
  }
  void *operator new(size_t sz) {
    return Allocator::allocSimple(sz);
  }
  void operator delete(void *p) {
    Allocator::free(p);
  }
  EntityOrigin();
  EntityOrigin(const ConstResourcePointer<Entity> &);
  EntityOrigin(const ConstResourcePointer<Entity> &,
	       const Location &refLocation);
  EntityOrigin(const ConstResourcePointer<Entity> &,
	       const Location &refLocation, Index refLength,
	       Origin::RefEndType refEndType, CString &origName,
	       int shortrefIndex = -1);
  ~EntityOrigin();
  EntityOrigin *copy() const;
  void noteCharRef(Index replacementIndex, const NamedCharRef &);
  const ConstResourcePointer<Entity> &entity() const { return entity_; }
  IndexType lookupIndex(Index ind, Offset &charOffset,
			NamedCharRef &ref) const;
  void setExternalInfo(ExternalInfo *);
  const Location &refLocation() const;
  const EntityOrigin *asEntityOrigin() const;
  const Location &parent() const;
  const ExternalInfo *externalInfo() const;
  Index refLength() const;
  Origin::RefEndType refEndType() const;
  Boolean shortref(int &) const;
  const CString &origName() const;
private:
  EntityOrigin(const EntityOrigin &); // undefined
  void operator=(const EntityOrigin &);	// undefined
  GrowableVector<EntityOriginNamedCharRef> charRefs_;
  CString charRefOrigNames_;
  Owner<ExternalInfo> externalInfo_; // 0 for internal entities
  ConstResourcePointer<Entity> entity_;	// 0 for document entity
  Location refLocation_;	// where referenced from
  // total length of reference
  // (characters that were replaced by the entity)
  Index refLength_;
  Origin::RefEndType refEndType_;
  CString origName_;
  int shortrefIndex_;
};

class NumericCharRefOrigin : public Origin {
public:
  NumericCharRefOrigin(const Location &start, Index endIndex,
		       Origin::RefEndType refEndType);
  const Location &parent() const;
  Index refLength() const;
  Origin::RefEndType refEndType() const;
  Boolean isNumericCharRef(Origin::RefEndType &refEndType) const;
private:
  Location start_;
  Index end_;
  Origin::RefEndType refEndType_;
};

// a delimiter specified in bracketed text

class BracketOrigin : public Origin {
public:
  enum Position { open, close };
  BracketOrigin(const Location &, Position);
  const Location &parent() const;
  Boolean inBracketedTextOpenDelim() const;
  Boolean inBracketedTextCloseDelim() const;
private:
  Position pos_;
  Location loc_;
};

class ReplacementOrigin : public Origin {
public:
  ReplacementOrigin(const Location &, Char origChar);
  const Location &parent() const;
  Boolean origChar(Char &) const;
private:
  Location loc_;
  Char origChar_;
};

inline
Index NamedCharRefBase::refStartIndex() const
{
  return refStartIndex_;
}

inline
Origin::RefEndType NamedCharRefBase::refEndType() const
{
  return refEndType_;
}

inline
const CString &NamedCharRef::origName() const
{
  return origName_;
}

inline
const ExternalInfo *EntityOrigin::externalInfo() const
{
  return externalInfo_.pointer();
}

inline
EntityOrigin *EntityOrigin::copy() const
{
  CString origName(origName_);
  return new EntityOrigin(entity_, refLocation_, refLength_, refEndType_,
			  origName, shortrefIndex_);
}

inline
Origin::RefEndType EntityOrigin::refEndType() const
{
  return refEndType_;
}

inline
const CString &EntityOrigin::origName() const
{
  return origName_;
}

inline
Boolean EntityOrigin::shortref(int &i) const
{
  if (shortrefIndex_ >= 0) {
    i = shortrefIndex_;
    return 1;
  }
  else
    return 0;
}

inline
Origin::RefEndType NumericCharRefOrigin::refEndType() const
{
  return refEndType_;
}

#endif /* not Location_INCLUDED */
