// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif
#include "config.h"
#include "Text.H"
#include "Entity.H"
// for memcmp()
#include <string.h>

Text::Text()
{
}

void Text::addChar(Char c, const Location &loc)
{
  if (items_.length() == 0
      || items_.last().type != TextItem::data
      || loc.origin().pointer() != items_.last().loc.origin().pointer()
      || loc.index() != (items_.last().loc.index()
			 + (chars_.length() - items_.last().index))) {
    items_.grow();
    items_.last().loc = loc;
    items_.last().type = TextItem::data;
    items_.last().index = chars_.length();
  }
  chars_ += c;
}

void Text::addChars(const Char *p, size_t length, const Location &loc)
{
  if (items_.length() == 0
      || items_.last().type != TextItem::data
      || loc.origin().pointer() != items_.last().loc.origin().pointer()
      || loc.index() != (items_.last().loc.index()
			 + (chars_.length() - items_.last().index))) {
    items_.grow();
    items_.last().loc = loc;
    items_.last().type = TextItem::data;
    items_.last().index = chars_.length();
  }
  chars_.append(p, length);
}

void Text::addSimple(TextItem::Type type, const Location &loc)
{
  items_.grow();
  items_.last().loc = loc;
  items_.last().type = type;
  items_.last().index = chars_.length();
}

void Text::ignoreChar(Char c, const Location &loc)
{
  items_.grow();
  items_.last().loc = loc;
  items_.last().type = TextItem::ignore;
  items_.last().c = c;
  items_.last().index = chars_.length();
}

void Text::ignoreLastChar()
{
  size_t lastIndex = chars_.length() - 1;
  for (size_t i = items_.length() - 1; items_[i].index > lastIndex; i--)
    ;
  // lastIndex >= items_[i].index
  if (items_[i].index != lastIndex) {
    items_.grow();
    i++;
    for (size_t j = items_.length() - 1; j > i; j--)
      items_[j] = items_[j - 1];
    items_[i].index = lastIndex;
    items_[i].loc = items_[i - 1].loc;
    items_[i].loc += lastIndex - items_[i - 1].index;
  }
  
  items_[i].c = chars_[chars_.length() - 1];
  items_[i].type = TextItem::ignore;
  for (size_t j = i + 1; j < items_.length(); j++)
    items_[j].index = lastIndex;
  chars_.setLength(chars_.length() - 1);
}

void Text::clear()
{
  chars_.clear();
  items_.clear();
}

Boolean Text::startDelimLocation(Location &loc) const
{
  if (items_.length() == 0 || items_[0].type != TextItem::startDelim)
    return 0;
  loc = items_[0].loc;
  return 1;
}

Boolean Text::endDelimLocation(Location &loc) const
{
  if (items_.length() == 0)
    return 0;
  switch (items_.last().type) {
  case TextItem::endDelim:
  case TextItem::endDelimA:
    break;
  default:
    return 0;
  }
  loc = items_.last().loc;
  return 1;
}

Boolean Text::delimType(Boolean &lita) const
{
  if (items_.length() == 0)
    return 0;
  switch (items_.last().type) {
  case TextItem::endDelim:
    lita = 0;
    return 1;
  case TextItem::endDelimA:
    lita = 1;
    return 1;
  default:
    break;
  }
  return 0;
}

TextItem::TextItem()
{
}

void Text::moveTo(Text &to)
{
  items_.moveTo(to.items_);
  chars_.moveTo(to.chars_);
}

EsisTextIter::EsisTextIter(const Text &text)
: ptr_(&text),
  itemIndex_(0)
{
}

Boolean EsisTextIter::next(Type &type, const Char *&ptr, size_t &length)
{
  const Entity *entity;
  return next(type, ptr, length, entity);
}

Boolean EsisTextIter::next(Type &type, const Char *&ptr, size_t &length,
			   const Entity *&sdataEntity)
{
  const GrowableVectorD<TextItem> &items = ptr_->items_;
  if (itemIndex_ >= items.length())
    return 0;
  const CString &chars = ptr_->chars_;
  size_t charIndex = items[itemIndex_].index;
  for (;;) {
    if (items[itemIndex_].type == TextItem::sdataEntityStart) {
      if (items[itemIndex_].index == charIndex) {
	sdataEntity = (items[itemIndex_].loc.origin()
		       ->asEntityOrigin()->entity().pointer());
	while (items[++itemIndex_].type != TextItem::sdataEntityEnd)
	  ;
	length = items[itemIndex_++].index - charIndex;
	ptr = chars.pointer() + charIndex;
	type = sdata;
      }
      else {
	ptr = chars.pointer() + charIndex;
	length = items[itemIndex_].index - charIndex;
	type = data;
      }
      break;
    }
    itemIndex_++;
    if (itemIndex_ >= items.length()) {
      if (charIndex >= chars.length())
	return 0;
      ptr = chars.pointer() + charIndex;
      length = chars.length() - charIndex;
      type = data;
      break;
    }
  }
  return 1;
}

TextIter::TextIter(const Text &text)
: ptr_(&text),
  itemIndex_(0)
{
}

Boolean TextIter::next(TextItem::Type &type, const Char *&ptr, size_t &length,
		       Location &loc)
{
  const GrowableVectorD<TextItem> &items = ptr_->items_;
  if (itemIndex_ >= items.length())
    return 0;
  type = items[itemIndex_].type;
  loc = items[itemIndex_].loc;
  if (type == TextItem::ignore) {
    ptr = &items[itemIndex_].c;
    length = 1;
  }
  else {
    const CString &chars = ptr_->chars_;
    size_t charsIndex = items[itemIndex_].index;
    ptr = chars.pointer() + charsIndex;
    if (itemIndex_ + 1 < items.length())
      length = items[itemIndex_ + 1].index - charsIndex;
    else
      length = chars.length() - charsIndex;
  }
  itemIndex_++;
  return 1;
}

void Text::insertChars(const CString &s, const Location &loc)
{
  chars_.insert(0, s);
  items_.grow();
  for (size_t i = items_.length() - 1; i > 0; i--) {
    items_[i] = items_[i - 1];
    items_[i].index += s.length();
  }
  items_[0].loc = loc;
  items_[0].type = TextItem::data;
  items_[0].index = 0;
}

size_t Text::nDataEntities() const
{
  size_t n = 0;
  for (size_t i = 0; i < items_.length(); i++)
    if (items_[i].type == TextItem::cdataEntityStart
	|| items_[i].type == TextItem::sdataEntityStart)
      n++;
  return n;
}

// This is used to determine for a FIXED CDATA attribute
// whether a specified value if equal to the default value.

Boolean Text::fixedEqual(const Text &text) const
{
  EsisTextIter i1(*this);
  EsisTextIter i2(text);
  for (;;) {
    EsisTextIter::Type type1, type2;
    const Char *p1, *p2;
    size_t n1, n2;
    if (!i1.next(type1, p1, n1))
      return !i2.next(type2, p2, n2);
    if (!i2.next(type2, p2, n2)
	|| type1 != type2
	|| n1 != n2
	|| memcmp(p1, p2, n1*sizeof(Char)) != 0)
      return 0;
  }
  return 1;
}

Location Text::charLocation(size_t ind) const
{
  // Find the last item whose index <= ind.
  // Invariant:
  // indexes < i implies index <= ind
  // indexes >= lim implies index > ind
  // The first item will always have index 0.
  size_t i = 1;
  size_t lim = items_.length();
  while (i < lim) {
    size_t mid = i + (lim - i)/2;
    if (items_[mid].index > ind)
      lim = mid;
    else
      i = mid + 1;
  }
#if 0
  for (size_t i = 1; i < items_.length(); i++)
    if (items_[i].index > ind)
      break;
#endif
  i--;
  Location loc;
  // If items_.length() == 0, then i == lim.
  if (i < lim) {
    loc = items_[i].loc;
    loc += ind - items_[i].index;
  }
  return loc;
}
