// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifndef Text_INCLUDED
#define Text_INCLUDED 1
#ifdef __GNUG__
#pragma interface
#endif

#include "types.H"
#include "CString.H"
#include "CopyGrowableVectorD.H"
#include "Location.H"
#include <stddef.h>

struct TextItem {
  TextItem();
  enum Type {
    data,
    entityStart,
    entityEnd,
    // Have to have separate start and end for tokenized attribute values.
    sdataEntityStart,
    sdataEntityEnd,
    cdataEntityStart,
    cdataEntityEnd,
    startDelim,
    endDelim,
    endDelimA,
    ignore
    };
  Type type;
  // char that was ignored
  Char c;
  // location of this item
  // data - location of first char
  // (c/sdata)entityStart - location of first char of entity
  // (c/sdata)entityEnd - location of entity end in entity
  // ignore - location of ignored character
  // startDelim - location of first char of delimiter
  // endDelim(A) - location of first char of delimiter
  Location loc;
  // index of character in chars_ to which this applies
  size_t index;
};

// This is used to represent literals and attribute values.

class Text {
public:
  Text();
  void clear();
  void moveTo(Text &to);
  void addChar(Char c, const Location &);
  void addChars(const CString &, const Location &);
  void addChars(const Char *, size_t, const Location &);
  void insertChars(const CString &, const Location &);
  void ignoreChar(Char, const Location &);
  void ignoreLastChar();
  void addEntityStart(const Location &);
  void addEntityEnd(const Location &);
  void addCdataEntityStart(const Location &);
  void addCdataEntityEnd(const Location &);
  void addSdataEntityStart(const Location &);
  void addSdataEntityEnd(const Location &);
  void addStartDelim(const Location &loc);
  void addEndDelim(const Location &loc, Boolean lita);
  Location charLocation(size_t i) const;
  size_t length() const;
  Char lastChar() const;
  const CString &string() const;
  size_t nDataEntities() const;
  Boolean fixedEqual(const Text &) const;
  // This information is kept only for system ids and minimum literals.
  // Location of first char of start delimiter.
  Boolean startDelimLocation(Location &) const;
  // Location of first char of end delimiter
  Boolean endDelimLocation(Location &) const;
  // Is delimiter a lit or lita?
  Boolean delimType(Boolean &lita) const;
private:
  void addSimple(TextItem::Type, const Location &);
  CString chars_;
  CopyGrowableVectorD<TextItem> items_;
  friend class EsisTextIter;
  friend class TextIter;
};

class Entity;

class EsisTextIter {
public:
  EsisTextIter(const Text &);
  enum Type {
    data,
    sdata
    };
  Boolean next(Type &, const Char *&, size_t &);
  Boolean next(Type &, const Char *&, size_t &,
	       const Entity *&sdataEntity);
private:
  const Text *ptr_;
  size_t itemIndex_;
};

class TextIter {
public:
  TextIter(const Text &);
  Boolean next(TextItem::Type &, const Char *&, size_t &,
	       Location &);
private:
  const Text *ptr_;
  size_t itemIndex_;
};

inline
size_t Text::length() const
{
  return chars_.length();
}

inline
Char Text::lastChar() const
{
  return chars_[chars_.length() - 1];
}

inline
const CString &Text::string() const
{
  return chars_;
}

inline
void Text::addEntityStart(const Location &loc)
{
  addSimple(TextItem::entityStart, loc);
}

inline
void Text::addEntityEnd(const Location &loc)
{
  addSimple(TextItem::entityEnd, loc);
}

inline
void Text::addCdataEntityStart(const Location &loc)
{
  addSimple(TextItem::cdataEntityStart, loc);
}

inline
void Text::addCdataEntityEnd(const Location &loc)
{
  addSimple(TextItem::cdataEntityEnd, loc);
}

inline
void Text::addSdataEntityStart(const Location &loc)
{
  addSimple(TextItem::sdataEntityStart, loc);
}

inline
void Text::addSdataEntityEnd(const Location &loc)
{
  addSimple(TextItem::sdataEntityEnd, loc);
}

inline
void Text::addChars(const CString &s, const Location &loc)
{
  addChars(s.pointer(), s.length(), loc);
}

inline
void Text::addStartDelim(const Location &loc)
{
  addSimple(TextItem::startDelim, loc);
}

inline
void Text::addEndDelim(const Location &loc, Boolean lita)
{
  addSimple(lita ? TextItem::endDelimA : TextItem::endDelim,
	    loc);
}

#endif /* not Text_INCLUDED */
