// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#include "config.h"
#include "SgmlParser.H"
#include "Event.H"
#include "MessageReporter.H"
#include "RastEventHandler.H"
#include "OutputCharStream.H"
#include "EasyParser.H"
#include "Boolean.H"
#include "programName.H"
#include "macros.H"
#include "getopt.h"
#include "ErrnoMessageArg.H"
#include "RastMessages.H"

#include "outOfMemory.H"
#include "xnew.h"

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <errno.h>

static Boolean exotericaMode = 0;

inline
void usage(Messenger &mgr)
{
  mgr.message(RastMessages::usageInfo);
}

class XRastEventHandler : public RastEventHandler {
public:
  XRastEventHandler(SgmlParser *,
		    const char *filename,
		    const OutputCodingSystem *,
		    const InputCodingSystem *,
		    MessageReporter *reporter);
  void message(MessageEvent *);
  void truncateOutput();
private:
  MessageReporter *reporter_;
  // file_ must come before os_ so it gets inited first
  filebuf file_;
  IosOutputCharStream os_;
  const char *filenameBytes_;
  const InputCodingSystem *filenameCodingSystem_;
};

int main(int argc, char **argv)
{
  set_new_handler(outOfMemory);
  programName = argv[0];
  int opt;
  char arg[2];
  EasyParser easy;
  
  opterr = 0;			// no messages from getopt
  // Colon at beginning is Posix.2ism that says to return : rather than ? for
  // missing option argument.
  while ((opt = getopt(argc, argv, ":ewxc:")) >= 0) {
    switch (opt) {
    case 'c':
      easy.addCatalogSysid(optarg);
      break;
    case 'e':
      // describe open entities in error messages
      easy.addOption(MessageReporter::openEntities);
      break;
    case 'x':
      exotericaMode = 1;
      break;
    case 'w':
      easy.options.warnShould = 1;
      easy.options.warnUndefinedElement = 1;
      easy.options.warnSgmlDecl = 1;
      easy.options.warnDuplicateEntity = 1;
      easy.options.warnDefaultEntityReference = 1;
      break;
    case ':':
      arg[0] = optopt;
      arg[1] = '\0';
      easy.message(RastMessages::missingOptionArgError,
		   StringMessageArg(easy.codingSystem()->convertIn(arg)));
      usage(easy);
      return 1;
    case '?':
      arg[0] = optopt;
      arg[1] = '\0';
      easy.message(RastMessages::invalidOptionError,
		   StringMessageArg(easy.codingSystem()->convertIn(arg)));
      usage(easy);
      return 1;
    default:
      CANNOT_HAPPEN();
    }
  }
  if (argc - optind < 2) {
    easy.message(RastMessages::missingOperandError);
    usage(easy);
    return 1;
  }
  easy.init(argc - optind - 1, argv + optind);
  SgmlParser &parser = easy.parser();
  XRastEventHandler handler(&parser, argv[argc - 1], easy.codingSystem(),
			    easy.codingSystem(), &easy);
  parser.parseAll(handler);
  handler.end();
  return handler.hadError();
}

XRastEventHandler::XRastEventHandler(SgmlParser *parser,
				     const char *filenameBytes,
				     const OutputCodingSystem *codingSystem,
				     const InputCodingSystem *filenameCodingSystem,
				     MessageReporter *reporter)
: RastEventHandler(parser, reporter),
  reporter_(reporter),
  filenameBytes_(filenameBytes),
  filenameCodingSystem_(filenameCodingSystem)
{
  errno = 0;
  if (!file_.open(filenameBytes, ios::out|ios::trunc)) {
    reporter->message(RastMessages::cannotOpenOutputError,
		      StringMessageArg(filenameCodingSystem_
				       ->convertIn(filenameBytes_)),
		      ErrnoMessageArg(errno));
    exit(1);
  }
  os_.open(&file_, codingSystem);
  setOutputStream(&os_);
}

void XRastEventHandler::truncateOutput()
{
  os_.flush();
  errno = 0;
  if (!file_.close())
    reporter_->message(RastMessages::closeOutputError,
		       StringMessageArg(filenameCodingSystem_
					->convertIn(filenameBytes_)),
		       ErrnoMessageArg(errno));
  errno = 0;
  if (!file_.open(filenameBytes_, ios::out|ios::trunc)) {
    reporter_->message(RastMessages::cannotOpenOutputError,
		       StringMessageArg(filenameCodingSystem_
					->convertIn(filenameBytes_)),
		       ErrnoMessageArg(errno));
    exit(1);
  }
}

void XRastEventHandler::message(MessageEvent *event)
{
  static const char msg[] =
    "character %1 is not significant in the reference concrete syntax"
    " and so cannot occur in a literal in the SGML declaration"
    " except as the replacement of a character reference";
  // FIXME
  if (exotericaMode
      && strcmp(event->message().type->id, msg) == 0) {
    delete event;
    return;
  }
  reporter_->reportMessage(event->message());
  RastEventHandler::message(event);
}
