// Copyright (c) 1995 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif

#include "config.h"
#include "CopyEventHandler.H"
#include "macros.H"

#ifdef __GNUG__
inline
#endif
Boolean CopyEventHandler::doNothing(Event *event)
{
  if ((normalizeFlags_ & normalizeExpand)
      ? entityLevel_ < outputEntityLevel_
      : entityLevel_ != outputEntityLevel_) {
    delete event;
    return 1;
  }
  else
    return 0;
}

CopyEventHandler::CopyEventHandler(OutputCharStream *os,
				   unsigned normalizeFlags,
				   const CString &outputEntity)
: os_(os), topOs_(os), inInstance_(0), entityLevel_(0),
  normalizeFlags_(normalizeFlags), outputEntity_(outputEntity),
  omittagHoist_(0), inSpecialMarkedSection_(0)
{
  outputEntityLevel_ = outputEntity_.length() == 0 ? 0 : unsigned(-1);
}

void CopyEventHandler::sgmlDecl(SgmlDeclEvent *event)
{
  syntax_ = event->instanceSyntaxPointer();
  sd_ = event->sdPointer();
  delete event;
}

void CopyEventHandler::endProlog(EndPrologEvent *event)
{
  inInstance_ = 1;
  delete event;
}

void CopyEventHandler::data(DataEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  if (event->entity())
    entityRef(event->location().origin()->asEntityOrigin());
  else {
    size_t n = event->dataLength();
    unsigned long dummy;
    if (n > 1 || !event->isRe(dummy))
      writeData(event->data(), n, event->location());
  }
  delete event;
}

void CopyEventHandler::writeData(const Char *p, size_t n,
				 const Location &loc)
{
  Origin::RefEndType refEndType;
  if (n == 1 && loc.origin()->isNumericCharRef(refEndType))
    numCharRef(*p, refEndType, loc.origin()->refLength());
  else
    withNamedCharRef(p, n, loc);
}

void CopyEventHandler::withNamedCharRef(const Char *p, size_t n,
					const Location &loc)
{
  if (n > 0) {
    const Origin *origin = loc.origin().pointer();
    if (origin) {
      const EntityOrigin *entityOrigin = origin->asEntityOrigin();
      if (entityOrigin) {
	NamedCharRef ref;
	Offset off;
	if (entityOrigin->lookupIndex(loc.index(), off, ref)
	    == EntityOrigin::replacementCharIndex) {
	  const CString &delim = syntax_->delimGeneral(Syntax::dCRO);
	  withNamedCharRef(delim.pointer(), delim.length(),
			   Location(loc.origin(), ref.refStartIndex()));
	  os() << ref.origName();
	  refEnd(ref.refEndType());
	  p++;
	  n--;
	}
      }
    }
  }
  os().write(p, n);
}

void CopyEventHandler::reOrigin(ReOriginEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  Char c = event->re();
  withNamedCharRef(&c, 1, event->location());
  delete event;
}

void CopyEventHandler::sSep(SSepEvent *event)
{
  if (doNothing(event))
    return;
  withNamedCharRef(event->data(), event->dataLength(), event->location());
  delete event;
}

void CopyEventHandler::ignoredRs(IgnoredRsEvent *event)
{
  if (doNothing(event))
    return;
  Char c = event->rs();
  withNamedCharRef(&c, 1, event->location());
  delete event;
}

void CopyEventHandler::startElement(StartElementEvent *event)
{
  ElementEvent::Minimization min = event->minimization();
  if (min == ElementEvent::omittedOmittag) {
    if (normalizeFlags_ & normalizeExpand) {
      if (outputEntityLevel_ > entityLevel_ - omittagHoist_) {
	delete event;
	return;
      }
      if (omittagHoist_ >= entityStack_.length())
	os_ = topOs_;
    }
    else if (entityLevel_ - omittagHoist_ != outputEntityLevel_) {
      delete event;
      return;
    }
  }
  else {
    omittagHoist_ = 0;
    if (doNothing(event))
      return;
  }
  if (normalizeFlags_ & normalizeExpandAll)
    handleChange();
  if (min != ElementEvent::omittedOmittag) {
    os() << syntax_->delimGeneral(Syntax::dSTAGO);
    if (min != ElementEvent::empty) {
      const ElementType *elementType = event->elementType();
      if (elementType->index() >= elementTypeOrigNames_.length())
	elementTypeOrigNames_.setLength(elementType->index() + 1);
      CString &elementTypeOrigName
	= elementTypeOrigNames_[elementType->index()];
      if (elementTypeOrigName.length() == 0) {
	elementTypeOrigName = event->origName();
	// add rank if missing
	elementTypeOrigName.append(event->name().pointer() +
				   elementTypeOrigName.length(),
				   event->name().length()
				   - elementTypeOrigName.length());
      }
      os() << event->origName();
      if (normalizeFlags_ & normalizeRank) {
	for (size_t i = event->origName().length();
	     i < event->name().length();
	     i++) {
	  handleChange();
	  os().put(event->name()[i]);
	}
      }
      const AttributeList &atts = event->attributes();
      size_t nAtt = atts.length();
      unsigned *order = new unsigned[nAtt];
      unsigned i;
      for (i = 0; i < nAtt; i++)
	order[i] = unsigned(-1);
      for (i = 0; i < nAtt; i++)
	if (atts.specified(i))
	  order[atts.spec(i)->specIndex()] = i;
      for (i = 0; i < nAtt; i++)
	if (order[i] != unsigned(-1)) {
	  const AttributeSpec *spec = atts.spec(order[i]);
	  const Text *text;
	  if (!atts.value(order[i])->text(text))
	    CANNOT_HAPPEN();
	  os() << spec->initialS();
	  CString origName = spec->origName();
	  if (origName.length() > 0) {
	    os() << origName << spec->preViS()
	      << syntax_->delimGeneral(Syntax::dVI) << spec->postViS();
	  }
	  else if (normalizeFlags_ & (normalizeAttname | normalizeAttvalue)) {
	    handleChange();
	    os() << atts.name(i) << syntax_->delimGeneral(Syntax::dVI);
	  }
	  attributeValue(*text);
	}
      if (normalizeFlags_ & normalizeAttspec) {
	for (i = 0; i < nAtt; i++) {
	  const Text *text;
	  if (!atts.specified(i) && atts.value(i)->text(text)) {
	    handleChange();
	    os().put(syntax_->standardFunction(Syntax::fSPACE));
	    os() << atts.name(i) << syntax_->delimGeneral(Syntax::dVI);
	    attributeValue(*text);
	  }
	}
      }
      delete [] order;
      os() << event->finalS();
      if (min == ElementEvent::complete)
	os() << syntax_->delimGeneral(Syntax::dTAGC);
      else if (min == ElementEvent::net && !(normalizeFlags_ & normalizeNet))
	os() << syntax_->delimGeneral(Syntax::dNET);
      else if (min != ElementEvent::unclosed
	       || (normalizeFlags_ && normalizeUnclosed)) {
	handleChange();
	os() << syntax_->delimGeneral(Syntax::dTAGC);
      }
    }
    else {
      Boolean isEmpty = 1;
      if (normalizeFlags_ & normalizeEmptytag) {
	handleChange();
	os() << elementTypeOrigName(event->elementType());
	isEmpty = 0;
      }
      if (normalizeFlags_ & normalizeAttspec) {
	const AttributeList &atts = event->attributes();
	size_t nAtt = atts.length();
	for (unsigned i = 0; i < nAtt; i++) {
	  const Text *text;
	  if (atts.value(i)->text(text)) {
	    handleChange();
	    if (isEmpty) {
	      os() << elementTypeOrigName(event->elementType());
	      isEmpty = 0;
	    }
	    os().put(syntax_->standardFunction(Syntax::fSPACE));
	    os() << atts.name(i) << syntax_->delimGeneral(Syntax::dVI);
	    attributeValue(*text);
	  }
	}
      }
      os() << syntax_->delimGeneral(Syntax::dTAGC);
    }
  }
  else if (normalizeFlags_ & normalizeOmittag) {
    if (inSpecialMarkedSection_) {
      reportTagInSpecialMarkedSection(event->location());
      return;
    }
    handleChange();
    os() << syntax_->delimGeneral(Syntax::dSTAGO)
      << elementTypeOrigName(event->elementType());
    if (normalizeFlags_ & normalizeAttspec) {
      const AttributeList &atts = event->attributes();
      size_t nAtt = atts.length();
      for (unsigned i = 0; i < nAtt; i++) {
	const Text *text;
	if (atts.value(i)->text(text)) {
	  os().put(syntax_->standardFunction(Syntax::fSPACE));
	  os() << atts.name(i) << syntax_->delimGeneral(Syntax::dVI);
	  attributeValue(*text);
	}
      }
    }
    os() << syntax_->delimGeneral(Syntax::dTAGC);
  }
  delete event;
  if (entityStack_.length() > 0 && os_ == topOs_)
    os_ = &entityStack_.last().str;
}

void CopyEventHandler::attributeValue(const Text &text)
{
  TextIter iter(text);
  TextItem::Type type;
  const Char *p;
  size_t n;
  Location loc;
  CString delim;
  Boolean lita;
  if (text.delimType(lita))
    delim = syntax_->delimGeneral(lita ? Syntax::dLITA : Syntax::dLIT);
  else if (normalizeFlags_ & normalizeAttvalue) {
    handleChange();
    delim = syntax_->delimGeneral(Syntax::dLIT);
  }
  os() << delim;
  int level = 0;
  while (iter.next(type, p, n, loc)) {
    switch (type) {
    case TextItem::data:
      if (!level) {
	Char orig;
	if (n == 1 && loc.origin()->origChar(orig))
	  writeData(&orig, 1, loc.origin()->parent());
	else
	  writeData(p, n, loc);
      }
      break;
    case TextItem::entityStart:
    case TextItem::cdataEntityStart:
    case TextItem::sdataEntityStart:
      if (!level++)
	entityRef(loc.origin()->asEntityOrigin());
      break;
    case TextItem::entityEnd:
    case TextItem::cdataEntityEnd:
    case TextItem::sdataEntityEnd:
      level--;
      break;
    case TextItem::startDelim:
    case TextItem::endDelim:
    case TextItem::endDelimA:
      break;
    case TextItem::ignore:
      if (!level)
	writeData(p, n, loc);
      break;
    }
  }
  os() << delim;
}

void CopyEventHandler::endElement(EndElementEvent *event)
{
  if (event->minimization() == ElementEvent::omittedOmittag) {
    if (normalizeFlags_ & normalizeExpand) {
      if (outputEntityLevel_ > entityLevel_ - omittagHoist_) {
	delete event;
	return;
      }
      if (omittagHoist_ >= entityStack_.length())
	os_ = topOs_;
    }
    else if (entityLevel_ - omittagHoist_ != outputEntityLevel_) {
      delete event;
      return;
    }
  }
  else {
    omittagHoist_ = 0;
    if (doNothing(event))
      return;
  }
  if (normalizeFlags_ & normalizeExpandAll)
    handleChange();
  switch (event->minimization()) {
  case ElementEvent::unclosed:
  case ElementEvent::complete:
    os() << syntax_->delimGeneral(Syntax::dETAGO);
    if (normalizeFlags_ && normalizeRank) {
      os() << event->origName();
      for (size_t i = event->origName().length();
	   i < event->name().length();
	   i++) {
	handleChange();
	os().put(event->name()[i]);
      }
      os() << event->finalS();
    }
    else
      os() << event->origChars();
    if (event->minimization() == ElementEvent::complete)
      os() << syntax_->delimGeneral(Syntax::dTAGC);
    else if (normalizeFlags_ & normalizeUnclosed) {
      handleChange();
      os() << syntax_->delimGeneral(Syntax::dTAGC);
    }
    break;
  case ElementEvent::empty:
    os() << syntax_->delimGeneral(Syntax::dETAGO);
    if (normalizeFlags_ & normalizeEmptytag) {
      handleChange();
      os() << elementTypeOrigName(event->elementType());
    }
    os() << syntax_->delimGeneral(Syntax::dTAGC);
    break;
  case ElementEvent::net:
    if (normalizeFlags_ & normalizeNet) {
      handleChange();
      os() << syntax_->delimGeneral(Syntax::dETAGO)
	<< elementTypeOrigName(event->elementType())
	  << syntax_->delimGeneral(Syntax::dTAGC);
    }
    else
      os() << syntax_->delimGeneral(Syntax::dNET);
    break;
  case ElementEvent::omittedOmittag:
    if (normalizeFlags_ & normalizeOmittag) {
      if (inSpecialMarkedSection_) {
	reportTagInSpecialMarkedSection(event->location());
	return;
      }
      handleChange();
      os() << syntax_->delimGeneral(Syntax::dETAGO)
	<< elementTypeOrigName(event->elementType())
	  << syntax_->delimGeneral(Syntax::dTAGC);
    }
    break;
  case ElementEvent::omittedEmpty:
  case ElementEvent::omittedConref:
    break;
  }
  delete event;
  if (entityStack_.length() > 0 && os_ == topOs_)
    os_ = &entityStack_.last().str;
}

void CopyEventHandler::pi(PiEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  if (event->entity())
    entityRef(event->location().origin()->asEntityOrigin());
  else {
    os() << syntax_->delimGeneral(Syntax::dPIO);
    os().write(event->data(), event->dataLength());
    os() << syntax_->delimGeneral(Syntax::dPIC);
  }
  delete event;
}

void CopyEventHandler::sdataEntity(SdataEntityEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  entityRef(event->location().origin()->asEntityOrigin());
  delete event;
}

void CopyEventHandler::externalDataEntity(ExternalDataEntityEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  entityRef(event->entityOrigin().pointer());
  delete event;
}

void CopyEventHandler::subdocEntity(SubdocEntityEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  entityRef(event->entityOrigin().pointer());
  delete event;
}

void CopyEventHandler::commentDecl(CommentDeclEvent *event)
{
  if (doNothing(event))
    return;
  if (inInstance_)
    declaration(event->decl());
  delete event;
}

void CopyEventHandler::markedSectionStart(MarkedSectionStartEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  declaration(event->decl());
#if 0
  os() << syntax_->delimGeneral(Syntax::dMDO)
    << syntax_->delimGeneral(Syntax::dDSO);
  switch (event->status()) {
  case MarkedSectionEvent::include:
    os() << syntax_->reservedName(Syntax::rINCLUDE);
    break;
  case MarkedSectionEvent::rcdata:
    os() << syntax_->reservedName(Syntax::rRCDATA);
    break;
  case MarkedSectionEvent::cdata:
    os() << syntax_->reservedName(Syntax::rCDATA);
    break;
  case MarkedSectionEvent::ignore:
    os() << syntax_->reservedName(Syntax::rIGNORE);
    break;
  }
  os() << syntax_->delimGeneral(Syntax::dDSO);
#endif
  switch (event->status()) {
  case MarkedSectionEvent::rcdata:
  case MarkedSectionEvent::cdata:
    inSpecialMarkedSection_ = 1;
    break;
  default:
    break;
  }
  delete event;
}

void CopyEventHandler::markedSectionEnd(MarkedSectionEndEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  os() << syntax_->delimGeneral(Syntax::dMSC)
    << syntax_->delimGeneral(Syntax::dMDC);
  inSpecialMarkedSection_ = 0;
  delete event;
}

void CopyEventHandler::ignoredChars(IgnoredCharsEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  os().write(event->data(), event->dataLength());
  delete event;
}

void CopyEventHandler::usemap(UsemapEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  declaration(event->decl());
#if 0
  os() << syntax_->delimGeneral(Syntax::dMDO)
    << syntax_->reservedName(Syntax::rUSEMAP);
  os().put(syntax_->standardFunction(Syntax::fSPACE));
  const CString &name = event->map()->name();
  if (name.length() == 0)
    os() << syntax_->delimGeneral(Syntax::dRNI)
      << syntax_->reservedName(Syntax::rEMPTY);
  else
    os() << name;
  os() << syntax_->delimGeneral(Syntax::dMDC);
#endif
  delete event;
}

void CopyEventHandler::uselink(UselinkEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  declaration(event->decl());
  delete event;
}

void CopyEventHandler::entityStart(EntityStartEvent *event)
{
  if (event->entity()->name() == outputEntity_)
    outputEntityLevel_ = entityLevel_ + 1;
  if (normalizeFlags_ & normalizeOmittagHoist) {
    if (event->entity()->asInternalEntity())
      omittagHoist_++;
    else
      omittagHoist_ = 0;
  }
  if (doNothing(event)) {
    entityLevel_++;
    return;
  }
  entityLevel_++;
  if ((normalizeFlags_ & normalizeExpand)
      && entityLevel_ > outputEntityLevel_) {
    entityStack_.grow().ref = event->entityOrigin();
    os_ = &entityStack_.last().str;
  }
  entityOrigin_ = event->entityOrigin();
  delete event;
}

void CopyEventHandler::entityEnd(EntityEndEvent *event)
{
  if (omittagHoist_ > 0)
    omittagHoist_--;
  if (entityLevel_-- == outputEntityLevel_) {
    outputEntityLevel_ = unsigned(-1);
    outputEntity_.clear();
  }
  else if (!(normalizeFlags_ & normalizeExpand)
	   && entityLevel_ == outputEntityLevel_) {
    entityRef(entityOrigin_.pointer());
    entityOrigin_.clear();
  }
  else if ((normalizeFlags_ & normalizeExpand)
	   && entityLevel_ >= outputEntityLevel_) {
    if (entityStack_.length() > 0) {
      ConstResourcePointer<EntityOrigin> origin
	= entityStack_.last().ref;
      entityStack_.setLength(entityStack_.length() - 1);
      if (entityStack_.length() > 0)
	os_ = &entityStack_.last().str;
      else
	os_ = topOs_;
      entityRef(origin.pointer());
    }
  }
  delete event;
}

void CopyEventHandler::declaration(const Decl &decl)
{
  int level = 0;
  os() << syntax_->delimGeneral(Syntax::dMDO);
  for (size_t i = 0; i < decl.length(); i++) 
    switch (decl[i].type) {
    case DeclItem::delimiter:
      if (!level)
	os() << syntax_->delimGeneral(decl[i].delimIndex);
      break;
    case DeclItem::indicatedReservedName:
      if (!level)
	os() << syntax_->delimGeneral(Syntax::dRNI);
      // fall through
    case DeclItem::reservedName:
    case DeclItem::name:
    case DeclItem::s:
      if (!level)
	os() << decl[i].str;
      break;
    case DeclItem::comment:
      if (!level)
	os() << syntax_->delimGeneral(Syntax::dCOM) << decl[i].str
	  << syntax_->delimGeneral(Syntax::dCOM);
      break;
    case DeclItem::entityStart:
      if (!level++)
	entityRef(decl[i].loc.origin()->asEntityOrigin());
      break;
    case DeclItem::entityEnd:
      level--;
      break;
    case DeclItem::other:
      CANNOT_HAPPEN();
    }
}

void CopyEventHandler::entityRef(const EntityOrigin *origin)
{
  int i;
  if (origin->shortref(i)) {
    if (normalizeFlags_ & normalizeShortref) {
      handleChange();
      Boolean containsRE = 0;
      Boolean containsRS = 0;
      const CString &ref = origin->origName();
      for (size_t i = 0; i < ref.length(); i++) {
	if (ref[i] == syntax_->standardFunction(Syntax::fRE))
	  containsRE = 1;
	else if (ref[i] == syntax_->standardFunction(Syntax::fRS))
	  containsRS = 1;
      }
      if (containsRS)
	os().put(syntax_->standardFunction(Syntax::fRS));
      os() << syntax_->delimGeneral(Syntax::dERO) << origin->entity()->name();
      if (containsRE)
	os().put(syntax_->standardFunction(Syntax::fRE));
      else
	os() << syntax_->delimGeneral(Syntax::dREFC);
    }
    else
      os() << origin->origName();
  }
  else {
    Boolean isParameter
      = origin->entity()->declType() == Entity::parameterEntity;
    os() << syntax_->delimGeneral(isParameter ? Syntax::dPERO : Syntax::dERO)
      << origin->origName();
    refEnd(origin->refEndType());
  }
}

void CopyEventHandler::numCharRef(unsigned long n,
				  Origin::RefEndType refEndType,
				  int refLength)
{
  refLength -= syntax_->delimGeneral(Syntax::dCRO).length();
  if (refEndType == Origin::endRE)
    refLength -= 1;
  else if (refEndType == Origin::endRefc)
    refLength -= syntax_->delimGeneral(Syntax::dREFC).length();
  os() << syntax_->delimGeneral(Syntax::dCRO);
  CString str;
  do {
    str += sd_->execToDoc('0' + n % 10);
    n /= 10;
  } while (n != 0);
  refLength -= str.length();
  while (refLength > 0) {
    os().put(sd_->execToDoc('0'));
    refLength--;
  }
  for (size_t i = str.length(); i > 0; i--)
    os().put(str[i - 1]);
  refEnd(refEndType);
}

void CopyEventHandler::refEnd(Origin::RefEndType refEndType)
{
  switch (refEndType) {
  case Origin::endOmitted:
    break;
  case Origin::endRE:
    os().put(syntax_->standardFunction(Syntax::fRE));
    break;
  case Origin::endRefc:
    os() << syntax_->delimGeneral(Syntax::dREFC);
    break;
  }
}

const CString &CopyEventHandler::elementTypeOrigName(const ElementType *type)
{
  if (type->index() < elementTypeOrigNames_.length()
      && elementTypeOrigNames_[type->index()].length() > 0)
    return elementTypeOrigNames_[type->index()];
  else
    return type->name();
}

void CopyEventHandler::handleChange()
{
  if (os_ != topOs_) {
    os_ = topOs_;
    for (size_t i = 0; i < entityStack_.length(); i++) {
      CString tem;
      entityStack_[i].str.flush();
      entityStack_[i].str.extractString(tem);
      os() << tem;
    }
    entityStack_.setLength(0);
  }
}
