// Copyright (c) 1995 James Clark
// See the file COPYING for copying permission.

#include "config.h"
#include "SgmlParser.H"
#include "Event.H"
#include "MessageReporter.H"
#include "OutputCharStream.H"
#include "EasyParser.H"
#include "Boolean.H"
#include "programName.H"
#include "macros.H"
#include "getopt.h"
#include "SpamMessages.H"
#include "LineOutputCodingSystem.H"
#include "CopyEventHandler.H"
#include "InputSource.H"

#include "outOfMemory.H"
#include "xnew.h"

#include <stdlib.h>
#include <iostream.h>

class XCopyEventHandler : public CopyEventHandler {
public:
  XCopyEventHandler(OutputCharStream *os, unsigned normalizeFlags,
		    const CString &outputEntity, InputSource *doc,
		    MessageReporter *);
  Boolean hadError() const;
  void message(MessageEvent *);
  void endProlog(EndPrologEvent *);
private:
  void reportTagInSpecialMarkedSection(const Location &);
  MessageReporter *reporter_;
  InputSource *doc_;
  OutputCharStream *os_;
  Boolean hadError_;
};

inline
Boolean XCopyEventHandler::hadError() const
{
  return hadError_;
}

inline
void usage(Messenger &mgr)
{
  mgr.message(SpamMessages::usageInfo);
}

const Char RE = '\r';
const Char RS = '\n';

static
struct {
  const char *name;
  unsigned flag;
} normalizeTable[] = {
  { "omittag", CopyEventHandler::normalizeOmittag },
  { "shortref", CopyEventHandler::normalizeShortref },
  { "net", CopyEventHandler::normalizeNet },
  { "emptytag", CopyEventHandler::normalizeEmptytag },
  { "unclosed", CopyEventHandler::normalizeUnclosed },
  { "attname", CopyEventHandler::normalizeAttname },
  { "attvalue", CopyEventHandler::normalizeAttvalue },
  { "attspec", CopyEventHandler::normalizeAttspec },
  { "shorttag", CopyEventHandler::normalizeShorttag },
  { "rank", CopyEventHandler::normalizeRank },
};

int main(int argc, char **argv)
{
  set_new_handler(outOfMemory);
#ifdef __GNUG__
  // cout is a performance disaster in libg++ unless we do this.
  ios::sync_with_stdio(0);
#endif
  programName = argv[0];
  int opt;
  char arg[2];
  EasyParser easy;
  unsigned normalizeFlags = 0;
  Boolean copyProlog = 0;
  CString outputEntity;
  
  opterr = 0;			// no messages from getopt
  // Colon at beginning is Posix.2ism that says to return : rather than ? for
  // missing option argument.
  while ((opt = getopt(argc, argv, ":i:m:o:ehpwc:x")) >= 0) {
    switch (opt) {
    case 'c':
      easy.addCatalogSysid(optarg);
      break;
    case 'e':
      // describe open entities in error messages
      easy.addOption(MessageReporter::openEntities);
      break;
    case 'h':
      // hoist omitted tags out of entities
      normalizeFlags |= CopyEventHandler::normalizeOmittagHoist;
      break;
    case 'i':
      // pretend that optarg is defined as INCLUDE
      easy.addInclude(optarg);
      break;
    case 'm':
      {
	Boolean found = 0;
	for (size_t i = 0; i < SIZEOF(normalizeTable); i++)
	  if (strcmp(normalizeTable[i].name, optarg) == 0) {
	    found = 1;
	    normalizeFlags |= normalizeTable[i].flag;
	    break;
	  }
	if (!found) {
	  easy.message(SpamMessages::invalidMarkupArgument,
		       StringMessageArg(easy.codingSystem()->convertIn(optarg)));
	  usage(easy);
	}
      }
      break;
    case 'o':
      outputEntity = easy.codingSystem()->convertIn(optarg);
      copyProlog = 0;
      break;
    case 'p':
      copyProlog = 1;
      outputEntity.clear();
      break;
    case 'w':
      easy.options.warnShould = 1;
      easy.options.warnUndefinedElement = 1;
      easy.options.warnSgmlDecl = 1;
      easy.options.warnDuplicateEntity = 1;
      easy.options.warnDefaultEntityReference = 1;
      break;
    case 'x':
      if (normalizeFlags & CopyEventHandler::normalizeExpand)
	normalizeFlags |= CopyEventHandler::normalizeExpandAll;
      else
	normalizeFlags |= CopyEventHandler::normalizeExpand;
      break;
    case ':':
      arg[0] = optopt;
      arg[1] = '\0';
      easy.message(SpamMessages::missingOptionArgError,
		   StringMessageArg(easy.codingSystem()->convertIn(arg)));
      usage(easy);
      return 1;
    case '?':
      arg[0] = optopt;
      arg[1] = '\0';
      easy.message(SpamMessages::invalidOptionError,
		   StringMessageArg(easy.codingSystem()->convertIn(arg)));
      usage(easy);
      return 1;
    default:
      CANNOT_HAPPEN();
    }
  }
  easy.options.eventsWanted.addEsisPlus();
  easy.init(argc - optind, argv + optind);
  InputSource *doc;
  if (copyProlog)
    doc = easy.openDocumentEntity(argc - optind, argv + optind);
  else
    doc = 0;
  SgmlParser &parser = easy.parser();
  LineOutputCodingSystem line(easy.codingSystem(), RS, RE, '\n');
  IosOutputCharStream os(cout.rdbuf(), &line);
  XCopyEventHandler handler(&os, normalizeFlags, outputEntity, doc, &easy);
  parser.parseAll(handler);
  return handler.hadError();
}

XCopyEventHandler::XCopyEventHandler(OutputCharStream *os,
				     unsigned normalizeFlags,
				     const CString &outputEntity,
				     InputSource *doc,
				     MessageReporter *reporter)
: reporter_(reporter), hadError_(0), doc_(doc), os_(os),
  CopyEventHandler(os, normalizeFlags, outputEntity)
{
}

void XCopyEventHandler::message(MessageEvent *event)
{
  reporter_->reportMessage(event->message());
  if (event->message().isError())
    hadError_ = 1;
  delete event;
}

void XCopyEventHandler::endProlog(EndPrologEvent *event)
{
  if (doc_) {
    Location loc(event->location());
    const EntityOrigin *origin = loc.origin()->asEntityOrigin();
    if (origin) {
      NamedCharRef ref;
      Offset off;
      if (origin->lookupIndex(loc.index(), off, ref)
	  == EntityOrigin::normalCharIndex
	  && origin->entity().isNull()) {
	for (; off > 0; --off) {
	  Xchar c = doc_->get(*reporter_);
	  if (c == InputSource::eE)
	    break;
	  os_->put(Char(c));
	}
      }
    }
    delete doc_;
    doc_ = 0;
  }
  CopyEventHandler::endProlog(event);
}

void XCopyEventHandler::reportTagInSpecialMarkedSection(const Location &loc)
{
  reporter_->setNextLocation(loc);
  reporter_->message(SpamMessages::tagInSpecialMarkedSection);
}

